#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<unistd.h>
#include<pthread.h>
#include"errno.h"
#include"mpi.h"

// tag -16 is used for allgather.
int
MPI_Allreduce (void *sendbuf, void *recvbuf, int count,
               MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{
  int nump, rank, i;
  int bm, nb, cb;
  int partner;
  void *buf1, *buf2;
  MPI_Status status;
  int *z;
  MPI_User_function *userop;

  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  if (rank == -1) {
    printf ("Wrong communicator in calling MPI_Allreduce\n");
    fflush (stdout);
    return 1;
  }

  nb = _USF_num_bits (nump - 1);
  bm = _USF_make_bm (nb);

  cb = 1;
  for (cb = 1; cb <= bm; cb <<= 1) {

    if (rank == (nump - 1)) {
      for (i = rank + 1; i < bm * 2; i++) {
        partner = i ^ cb;
        if ((partner >= rank) || ((i & (~(cb - 1))) >= nump))
          continue;
        MPI_Send (sendbuf, count, datatype, partner, -16, comm);
      }
    }

    partner = rank ^ cb;
    if (partner >= nump) {      // do a recv from last processor!
      if ((rank != (nump - 1))
          && ((partner & (~(cb - 1))) < nump)) {
        MPI_Recv (recvbuf, count, datatype, (nump - 1), -16, comm, &status);
      } else
        continue;
    } else {
      MPI_Sendrecv (sendbuf, count, datatype, partner, -16,
                    recvbuf, count, datatype, partner, -16, comm, &status);
    }

    if (cb == bm) {
      buf1 = recvbuf;
      buf2 = sendbuf;
    } else {
      buf1 = sendbuf;
      buf2 = recvbuf;
    }
    switch (op) {
    case MPI_MAX:
      _USF_solve_max (buf1, buf2, count, datatype);
      break;
    case MPI_MIN:
      _USF_solve_min (buf1, buf2, count, datatype);
      break;
    case MPI_SUM:
      _USF_solve_sum (buf1, buf2, count, datatype);
      break;
    case MPI_PROD:
      _USF_solve_prod (buf1, buf2, count, datatype);
      break;
    case MPI_LAND:
      _USF_solve_land (buf1, buf2, count, datatype);
      break;
    case MPI_BAND:
      _USF_solve_band (buf1, buf2, count, datatype);
      break;
    case MPI_LOR:
      _USF_solve_lor (buf1, buf2, count, datatype);
      break;
    case MPI_BOR:
      _USF_solve_bor (buf1, buf2, count, datatype);
      break;
    case MPI_LXOR:
      _USF_solve_lxor (buf1, buf2, count, datatype);
      break;
    case MPI_BXOR:
      _USF_solve_bxor (buf1, buf2, count, datatype);
      break;
    case MPI_MINLOC:
      _USF_solve_2loc (buf1, buf2, count, datatype, 1);
      break;
    case MPI_MAXLOC:
      _USF_solve_2loc (buf1, buf2, count, datatype, 0);
      break;
    default:
      if (op >= 120) {
        //This is a user defined opreation
        userop = Get_User_function (op);
        if (userop == NULL) {
          printf ("This operation isn't defined!\n");
          exit (1);
        } else {
          userop (buf2, buf1, &count, &datatype);
        }
      } else {
        printf ("This operation isn't defined!\n");
        exit (1);
      }
    }
  }
  return 0;
}
