#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<unistd.h>
#include"errno.h"
#include"mpi.h"

//Tag 21 is used for Alltoall
int
MPI_Alltoall (void *sendbuf, int sendcount, MPI_Datatype sendtype,
              void *recvbuf, int recvcount, MPI_Datatype recvtype,
              MPI_Comm comm)
{
  int nump, rank, i, j, rsize, ssize, recvind;
  MPI_Request *send_req;
  MPI_Request *recv_req;
  MPI_Status *status;

  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  if (rank == -1) {
    printf ("Wrong communicator in calling MPI_Allgather\n");
    fflush (stdout);
    return 1;
  }

  rsize = _USF_mpi_sizeof (recvtype);
  ssize = _USF_mpi_sizeof (sendtype);
  recv_req = (MPI_Request *) malloc ((nump - 1) * sizeof (MPI_Request));
  send_req = (MPI_Request *) malloc ((nump - 1) * sizeof (MPI_Request));
  status = (MPI_Status *) malloc ((nump - 1) * sizeof (MPI_Status));

  // First set the receives, then send
  for (i = 0, j = 0; i < nump; i++, j++) {
    if (rank == i) {
      j--;
      continue;
    }
    MPI_Irecv ((recvbuf + (i * recvcount * rsize)), recvcount,
               recvtype, i, -21, comm, &(recv_req[j]));

  }
  for (i = 0, j = 0; i < nump; i++, j++) {
    if (rank == i) {
      j--;
      continue;
    }
    MPI_Isend ((sendbuf + (i * sendcount * ssize)), sendcount,
               sendtype, i, -21, comm, &(send_req[j]));

  }
  if (recvbuf != sendbuf) {
    memcpy ((recvbuf + (rank * recvcount * rsize)),
            (sendbuf + (rank * sendcount * ssize)), sendcount * ssize);
  }
  MPI_Waitall ((nump - 1), recv_req, status);
  MPI_Waitall ((nump - 1), send_req, status);
  return 0;
}
