#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <pthread.h>
#include "errno.h"
#include "mpi.h"

#ifdef _USF_MPI_TCP_
int blocking_for_big_message[_USF_MAXPROCESSORS] = { 0 };
#define DB(a,b,c)
#ifdef DEBUG
#undef DB
#define DB(a,b,c) printf (a,b,c); fflush (stdout);
#endif

void
_USF_handle_recv_post (int dest, int source, int tag, int comm, void *buf,
                       int size, int multiple)
{
  _USF_message_header hdr;      /* The header for sends */
  int send_size;                /* The number of bytes sent */
  int recv_size;                /* The number of bytes sent */

  hdr.source = source;
  hdr.dest = dest;
  hdr.tag = tag;
  hdr.comm = comm;
  if (!(multiple)) {            /* Ready mode short message */
    /* Send send init */
    hdr.size = size;
    hdr.type = _USF_SHORTMES;
    _USF_send_all (dest, sizeof (_USF_message_header), &hdr,
                   "Header isn't transmitted in handle_recv_post");
    _USF_send_all (dest, size, buf, "Can't sent ready mode message");
  } else {                      /* First of multiple chunks */
    /* Send send init */
    hdr.size = _USF_MAXDATASIZE;
    hdr.type = _USF_FCHUNK;
    _USF_send_all (dest, sizeof (_USF_message_header), &hdr,
                   "Header isn't transmitted in handle_recv_post");
    _USF_send_all (dest, _USF_MAXDATASIZE, buf,
                   "Can't sent ready mode message");
    DB ("rank %d, handle recv post multip2 %d\n", source, multiple);
  }
  /* updatie mpi_request in handle_msg */
  return;
}

void
_USF_serve_sysreq (int source)
{
  int i;                        /* Loop variable */
  int nump;                     /* # of procs in COMM_WORLD */
  int rank;                     /* the rank in COMM_WORLD */
  _USF_message_header hdr;      /* The message header */
  _USF_request_node *req1;      /* The request node used for lookup */
  _USF_sysreq_node *node;       /* The sysreq node used to search */
  MPI_Comm_size (MPI_COMM_WORLD, &nump);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);

  /* Search for a recv_post in sysreq list that matches a synch
     (multiple chunks) send from the current processor */
  /* This function uses the internal data structure ofd sysrequest
     data sturucture */
  DB ("rank %d, serve sysreq\n", rank, 1);
  pthread_mutex_lock (&_USF_mut_sysreq);
  source = source % _USF_MAXPROCESSORS;
  for (node = _USF_sysreq_hashtable[source]; node != NULL; node = node->next) {
    DB ("rank %d, forloop\n", rank, 1);
    if ((source == node->source) && (node->type == _USF_RECEIVE)) {
      break;
    }
  }

  if (node == NULL) {
    pthread_mutex_unlock (&_USF_mut_sysreq);
    return;
  }
  pthread_mutex_lock (&_USF_mut_req);
  req1 = _USF_request_lookup (node->tag, source, _USF_SEND, node->comm);


  if (req1 == NULL) {
    pthread_mutex_unlock (&_USF_mut_sysreq);
    pthread_mutex_unlock (&_USF_mut_req);
    return;
  }
  /* Found the matching long send */
  hdr.source = rank;
  hdr.dest = source;
  hdr.tag = node->tag;
  hdr.comm = node->comm;
  /* Remove the sysreq */
  pool_return (_USF_sysreq_node_pool, (pool_item *)
               _USF_sysreq_remove (node->tag, source, _USF_RECEIVE,
                                   node->comm));
  pthread_mutex_unlock (&_USF_mut_sysreq);
  /* Start communication */
  DB ("rank %d, start to handle recv post in servsys\n", rank, 1);
  _USF_handle_recv_post (source, rank, req1->req->tag, req1->req->comm,
                         req1->req->buf, req1->req->size,
                         req1->req->multiple);
  blocking_for_big_message[source]--;
  send_post_account[req1->req->peer]++;
  if (req1->req->multiple) {
    req1->req->nchunk = 1;
    _USF_synchr_list[source] = req1->req;
  } else {
    req1->req->done = 1;
    /* Check if it was blocking !!! */
    pthread_mutex_lock (&_USF_mut_block);
    if (req1->req->blocking) {
      DB ("rank %d, wake up main\n", rank, 1);
      req1->req->end_of_transmission = 1;
      pthread_cond_signal (&_USF_cond_block);
    } else
      req1->req->end_of_transmission = 1;
    pthread_mutex_unlock (&_USF_mut_block);
  }

  pthread_mutex_unlock (&_USF_mut_req);
}

void
_USF_handle_msg (int source)
{
  int i;                        /* Loop variable */
  int nump;                     /* # of procs in COMM_WORLD */
  int rank;                     /* the rank in COMM_WORLD */
  int recv_size;                /* The size of received packet */
  _USF_message_header hdr;      /* The message header */
  int size;                     /* Used for ready mode receive */
  _USF_request_node *req1, *req_probe;  /* The request node used for lookup */
  _USF_sysreq_node *sysreq1;    /* The sysreq node used to insert a node */
  int isongcomm = 0;            /* Is there an ongoiing communication */
  int header_msg = 0;

  MPI_Comm_size (MPI_COMM_WORLD, &nump);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);


  //Read data from source node
  // Modified by Chaney 10/3
  if (_USF_conn_list[source].next_msg == _USF_HEADER_TYPE) {
    _USF_recv_all (source, sizeof (_USF_message_header), &hdr,
                   "Header isn't received for handle_msg");
    header_msg = 1;
  } else {
    memcpy (&hdr, &(_USF_conn_list[source].hdr),
            sizeof (_USF_message_header));
  }
  // End of change by Chaney
#ifdef DEBUG
  printf ("rank %d, Message: so %d de %d ta %d co %d si %d ty %d\n",
          rank, hdr.source, hdr.dest, hdr.tag, hdr.comm, hdr.size, hdr.type);
  fflush (stdout);
#endif
  switch (hdr.type) {
  case _USF_RECVPOST:
    /* These are removed from sysreq by MPI_(I)SEND calls */
    /* Look for a send request in request list */
    DB ("rank %d, Get a recv_post\n", rank, 1);
    pthread_mutex_lock (&_USF_mut_req);
    req1 = _USF_request_lookup (hdr.tag, hdr.source, _USF_SEND, hdr.comm);
    if (req1 != NULL) {
      DB ("rank %d, NONNULL \n", rank, 1);
      if (_USF_synchr_list[source] != NULL)
        isongcomm = 1;
    }
    DB ("rank %d, isong %d\n", rank, isongcomm);
    if ((req1 == NULL) || isongcomm) {
      sysreq1 = (_USF_sysreq_node *) pool_take (_USF_sysreq_node_pool);
      if (sysreq1 == NULL)
        DB ("rank %d Memory error", rank, 1);
      pthread_mutex_lock (&_USF_mut_sysreq);
      _USF_sysreq_insert (hdr.tag, hdr.source,
                          _USF_RECEIVE, hdr.comm, hdr.size, sysreq1);
      pthread_mutex_unlock (&_USF_mut_sysreq);
    } else {
      _USF_handle_recv_post (hdr.source,
                             hdr.dest, hdr.tag, hdr.comm, req1->req->buf,
                             req1->req->size, req1->req->multiple);
      /* update mpi_request */
      if (req1->req->posted == 0)
        ++send_post_account[hdr.source];
      if (req1->req->multiple) {
        req1->req->nchunk = 1;
        _USF_synchr_list[source] = req1->req;
      } else {
        req1->req->done = 1;
        /* Check if it was blocking !!! */
        pthread_mutex_lock (&_USF_mut_block);
        if (req1->req->blocking) {
          DB ("rank %d, wake up main\n", rank, 1);
          req1->req->end_of_transmission = 1;
          pthread_cond_signal (&_USF_cond_block);
        } else
          req1->req->end_of_transmission = 1;
        pthread_mutex_unlock (&_USF_mut_block);
      }
    }
    pthread_mutex_unlock (&_USF_mut_req);
    break;
  case _USF_SENDPOST:
    /* These are removed from sysreq by MPI_(I)RECV calls */
    /* Send a recv post if unsent recv post in mpi req list */
    DB ("rank %d Get a send_post\n", rank, 1);

    if (hdr.send_seq_num <= recv_post_account[hdr.source])
      break;

    pthread_mutex_lock (&_USF_mut_req);

    /* Continue with the operation */
    req1 = _USF_request_lookup_p (hdr.tag, _USF_RECEIVE, hdr.comm, 0);
    if (req1 == NULL) {         /* put in the sys req queue */
      //Received a send-post but there is no recv req in system queue
      DB ("rank %d SENDPOST NULL, put in sysreq Q\n", rank, 1);
      sysreq1 = (_USF_sysreq_node *) pool_take (_USF_sysreq_node_pool);
      pthread_mutex_lock (&_USF_mut_sysreq);
      _USF_sysreq_insert (hdr.tag, hdr.source, _USF_SEND, hdr.comm,
                          hdr.size, sysreq1);
      pthread_mutex_unlock (&_USF_mut_sysreq);

      /* check if there is a probe */
      req_probe = _USF_request_lookup_p (MPI_ANY_TAG, _USF_PROBE,
                                         MPI_COMM_WORLD, 0);
      if (req_probe != NULL) {
        pthread_mutex_lock (&_USF_mut_block);
        if (req_probe->req->blocking) {
          req_probe->req->end_of_transmission = 1;
          pthread_cond_signal (&_USF_cond_block);
        } else
          req_probe->req->end_of_transmission = 1;
        pthread_mutex_unlock (&_USF_mut_block);
      }

    } else {
      /* will match a MPI_ANY_SOURCE nonsent recv */
      /* Send a recv_post */
      if (req1->req->peer != MPI_ANY_SOURCE) {
        pthread_mutex_unlock (&_USF_mut_req);
        break;
      }
      DB ("rank %d SP will match a MPI_ANY_SOURCE\n", rank, 1);
      hdr.source = rank;
      hdr.dest = source;
      hdr.tag = req1->req->tag;
      hdr.comm = req1->req->comm;
      hdr.size = 0;
      hdr.type = _USF_RECVPOST;
      req1->req->peer = hdr.dest;       // not sure
      _USF_send_all (req1->req->peer, sizeof (_USF_message_header), &hdr,
                     "Packet isn't transmitted");
      req1->req->posted = 1;
      recv_post_account[req1->req->peer]++;
    }
    pthread_mutex_unlock (&_USF_mut_req);
    break;
  case _USF_FCHUNK:
    DB ("rank %d, FCHUNK\n", rank, 1);
    if (header_msg) {
      _USF_conn_list[source].next_msg = hdr.type;
      memcpy (&(_USF_conn_list[source].hdr), &hdr,
              sizeof (_USF_message_header));
      break;
    }
    pthread_mutex_lock (&_USF_mut_req);
    req1 =
      _USF_request_lookup_m (hdr.tag, hdr.source, _USF_RECEIVE, hdr.comm, 0);
    _USF_recv_all (hdr.source, _USF_MAXDATASIZE, req1->req->buf,
                   "Can't receive ready mode message");
    req1->req->multiple = 1;
    req1->req->nchunk = 1;
    pthread_mutex_unlock (&_USF_mut_req);
    hdr.source = rank;
    hdr.dest = source;
    hdr.size = 0;
    hdr.type = _USF_ACKCHUNK;
    _USF_send_all (source, sizeof (_USF_message_header), &hdr,
                   "Packet isn't transmitted");
    _USF_conn_list[source].next_msg = _USF_HEADER_TYPE;
    break;
  case _USF_MCHUNK:
    if (header_msg) {
      _USF_conn_list[source].next_msg = hdr.type;
      memcpy (&(_USF_conn_list[source].hdr), &hdr,
              sizeof (_USF_message_header));
      break;
    }
    pthread_mutex_lock (&_USF_mut_req);
    req1 =
      _USF_request_lookup_m (hdr.tag, hdr.source, _USF_RECEIVE, hdr.comm, 1);
    _USF_recv_all (hdr.source, _USF_MAXDATASIZE,
                   req1->req->buf + _USF_MAXDATASIZE * req1->req->nchunk,
                   "Can't receive ready mode message");
    req1->req->nchunk++;
    pthread_mutex_unlock (&_USF_mut_req);
    hdr.source = rank;
    hdr.dest = source;
    hdr.size = 0;
    hdr.type = _USF_ACKCHUNK;
    _USF_send_all (source, sizeof (_USF_message_header), &hdr,
                   "Packet isn't transmitted");
    _USF_conn_list[source].next_msg = _USF_HEADER_TYPE;
    break;
  case _USF_LCHUNK:
    if (header_msg) {
      _USF_conn_list[source].next_msg = hdr.type;
      memcpy (&(_USF_conn_list[source].hdr), &hdr,
              sizeof (_USF_message_header));
      break;
    }
    pthread_mutex_lock (&_USF_mut_req);
    req1 =
      _USF_request_lookup_m (hdr.tag, hdr.source, _USF_RECEIVE, hdr.comm, 1);
    _USF_recv_all (hdr.source, hdr.size,
                   req1->req->buf + _USF_MAXDATASIZE * req1->req->nchunk,
                   "Can't receive ready mode message");
    req1->req->size = _USF_MAXDATASIZE * req1->req->nchunk + hdr.size;
    req1->req->done = 1;
    DB ("rank %d, done receiving!\n", rank, 1);
    /* Check if it was blocking !!! */
    pthread_mutex_lock (&_USF_mut_block);
    if (req1->req->blocking) {
      DB ("rank %d, wake up main\n", rank, 1);
      req1->req->end_of_transmission = 1;
      pthread_cond_signal (&_USF_cond_block);
    } else
      req1->req->end_of_transmission = 1;
    pthread_mutex_unlock (&_USF_mut_block);
    pthread_mutex_unlock (&_USF_mut_req);
    _USF_conn_list[source].next_msg = _USF_HEADER_TYPE;
    DB ("rank %d, LCHUNK ends\n", rank, 1);
    break;
  case _USF_ACKCHUNK:
    pthread_mutex_lock (&_USF_mut_req);
    hdr.source = rank;
    hdr.dest = source;
    hdr.tag = _USF_synchr_list[source]->tag;
    hdr.comm = _USF_synchr_list[source]->comm;
    /* Check if it the final message to send out */
#ifdef DEBUG
    printf ("rank %d, --> %d %d \n", rank,
            ((_USF_synchr_list[source]->nchunk + 1) * _USF_MAXDATASIZE)
            , _USF_synchr_list[source]->size);
    fflush (stdout);
#endif
    if (((_USF_synchr_list[source]->nchunk + 1) *       /* MCHUNK */
         _USF_MAXDATASIZE) < _USF_synchr_list[source]->size) {
      hdr.size = _USF_MAXDATASIZE;
      hdr.type = _USF_MCHUNK;
      _USF_send_all (source, sizeof (_USF_message_header), &hdr,
                     "Header isn't transmitted in ACKCHU");
      _USF_send_all (source, _USF_MAXDATASIZE,
                     _USF_synchr_list[source]->buf
                     + (_USF_synchr_list[source]->nchunk * _USF_MAXDATASIZE)
                     , "Can't sent ACKCHU, MCHUNK");
      _USF_synchr_list[source]->nchunk++;
      pthread_mutex_unlock (&_USF_mut_req);
    } else {                    /* LCHUNK */
      hdr.size = _USF_synchr_list[source]->size -
        (_USF_synchr_list[source]->nchunk * _USF_MAXDATASIZE);
      hdr.type = _USF_LCHUNK;
      _USF_send_all (source, sizeof (_USF_message_header), &hdr,
                     "Header isn't transmitted in ACKCHU");
      _USF_send_all (source, hdr.size,
                     _USF_synchr_list[source]->buf
                     + (_USF_synchr_list[source]->nchunk * _USF_MAXDATASIZE)
                     , "Can't sent ACKCHU, FCHUNK");
      DB ("rank %d, !!! %d \n", rank, _USF_synchr_list[source]->blocking);
      _USF_synchr_list[source]->done = 1;
      /* Check if it was blocking !!! */
      pthread_mutex_lock (&_USF_mut_block);
      if (_USF_synchr_list[source]->blocking) {
        DB ("rank %d, wake up main\n", rank, 1);
        _USF_synchr_list[source]->end_of_transmission = 1;
        pthread_cond_signal (&_USF_cond_block); // we might have problem here
      } else
        _USF_synchr_list[source]->end_of_transmission = 1;
      pthread_mutex_unlock (&_USF_mut_block);
      DB ("rank %d, signal sent! \n", rank, 1);
      _USF_synchr_list[source] = NULL;
      /* Done transmitting the message serve the recv post
         requests that came during transmission */
      pthread_mutex_unlock (&_USF_mut_req);
      _USF_serve_sysreq (source);
    }
    DB ("rank %d, mut_req unlocked! \n", rank, 1);
    break;
  case _USF_SHORTMES:
    DB ("rank %d Get a short mes \n", rank, 1);
    if (header_msg) {
      _USF_conn_list[source].next_msg = hdr.type;
      memcpy (&(_USF_conn_list[source].hdr), &hdr,
              sizeof (_USF_message_header));
      if (hdr.size != 0)
        break;
    }
    pthread_mutex_lock (&_USF_mut_req);
    req1 =
      _USF_request_lookup_m (hdr.tag, hdr.source, _USF_RECEIVE, hdr.comm, 0);
    if (hdr.size > req1->req->size) {
      printf ("Incoming message is too big in ready mode\n");
      Error (ERF_RECV);
    }
    req1->req->size = hdr.size;
    _USF_recv_all (hdr.source, req1->req->size,
                   req1->req->buf, "Can't receive ready mode message");
    req1->req->done = 1;
    /* Check if it was blocking !!! */
    pthread_mutex_lock (&_USF_mut_block);
    if (req1->req->blocking) {
      DB ("rank %d, wake up main\n", rank, 1);
      req1->req->end_of_transmission = 1;
      pthread_cond_signal (&_USF_cond_block);
    } else
      req1->req->end_of_transmission = 1;
    pthread_mutex_unlock (&_USF_mut_block);
    pthread_mutex_unlock (&_USF_mut_req);
    _USF_conn_list[source].next_msg = _USF_HEADER_TYPE;
    DB ("rank %d short mes received\n", rank, 1);
    break;
  case _USF_COMMTERM:
    if (rank == 0) {
      _USF_finalize_list[source] = 1;
    } else {
      printf ("Comm_term mes received in handle mes rank %d\n", rank);
      fflush (stdout);
      exit (0);
    }
    DB ("rank %d commterm received\n", rank, 1);
    break;
  default:                     /* Error! */
    printf ("Unexpected message type in handle_msg!\n");
    fflush (stdout);
    Error (ERF_RECV);
    break;
  }
  return;
}


void
_USF_serve_mpi_requests (int rank, int nump)
{
  int i;                        /* Loop variable */
  _USF_message_header hdr;      /* The message header */
  _USF_request_node *curr;      /* The pointer used to traverse the list */
  MPI_Request_Int *req;         /* The pointer used to store MPI_Request_Int */
  _USF_sysreq_node *sysreq1;    /* The pointer used to retrieve sysreq */
  int send_size;                /* The number of bytes sent */
  int sysreq_flag;              /* Shows if there was a matching sysreq */


  pthread_mutex_lock (&_USF_mut_req);
  curr = _USF_request_hd;
  /* Traverse and serve all the requests */
  while (curr != NULL) {
    req = curr->req;
    DB ("rank %d s_mpi_req fst type %d\n", rank, req->type);
    /* continue if already completed */
    if ((req->done) || (req->posted) || (req->active == 0)) {
      DB ("rank %d found completed or posted or inactive\n", rank, 1);
      curr = curr->next;
      continue;
    }
    if (req->type == _USF_SEND) {       /* type send */
      pthread_mutex_lock (&_USF_mut_sysreq);

      /* Check for a matching sysreq recv post */
      sysreq1 = _USF_sysreq_lookup (req->tag, req->peer,
                                    _USF_RECEIVE, req->comm);
      DB ("rank %d serve mpi _USF_SEND request.\n", rank, 1);
      if (sysreq1 == NULL) {    /* send send_post */
        /* if it is a ready mode message send
         * immediately */
        DB ("rank %d Send a send_post\n", rank, 1);
        hdr.source = rank;
        hdr.dest = req->peer;
        hdr.tag = req->tag;
        hdr.comm = req->comm;
        hdr.size = req->size;
        hdr.type = _USF_SENDPOST;
        hdr.send_seq_num = ++send_post_account[req->peer];
        _USF_send_all (req->peer, sizeof (_USF_message_header), &hdr,
                       "Packet isn't transmitted");
        req->posted = 1;
        DB ("rank %d Sent a send_post!!!\n", rank, 1);
      } else {                  /* if there is a request send the message */
        if (_USF_synchr_list[req->peer] == NULL) {
          send_post_account[req->peer]++;
          DB ("rank %d Send the message2\n", rank, 1);
          /* only serve long communication if no ongoing long communication */
          _USF_handle_recv_post (sysreq1->source,
                                 rank, sysreq1->tag, sysreq1->comm,
                                 req->buf, req->size, req->multiple);
          /* update mpi_request */
          if (req->multiple) {
            req->posted = 1;
            req->nchunk = 1;
            _USF_synchr_list[req->peer] = req;
          } else {
            req->done = 1;
            /* Check if it was blocking !!! */
            pthread_mutex_lock (&_USF_mut_block);
            if (req->blocking) {
              DB ("rank %d, wake up main\n", rank, 1);
              req->end_of_transmission = 1;
              pthread_cond_signal (&_USF_cond_block);
            } else
              req->end_of_transmission = 1;
            pthread_mutex_unlock (&_USF_mut_block);
          }
          pool_return (_USF_sysreq_node_pool, (pool_item *)
                       _USF_sysreq_remove (req->tag, req->peer,
                                           _USF_RECEIVE, req->comm));
        } else
          blocking_for_big_message[req->peer]++;
      }
      pthread_mutex_unlock (&_USF_mut_sysreq);
    } else if (req->type == _USF_RECEIVE) {     /* type receive */
      DB ("rank %d serve _USF_RECEIVE request\n", rank, 1);
      pthread_mutex_lock (&_USF_mut_sysreq);
      /* Check for a matching sysreq send post */
      DB ("rank %d, req->peer %d\n", rank, req->peer);
      sysreq1 = _USF_sysreq_lookup (req->tag, req->peer,
                                    _USF_SEND, req->comm);
      /* If there is remove it */
      if (sysreq1 == NULL) {
        DB ("rank %d, NOT FOUND SYSREQ\n", rank, 1);
        sysreq_flag = 0;
      } else {
        DB ("rank %d, FOUND SYSREQ src %d\n", rank, sysreq1->source);
        sysreq_flag = 1;
        req->peer = sysreq1->source;
        pool_return (_USF_sysreq_node_pool, (pool_item *)
                     _USF_sysreq_remove (req->tag, req->peer, _USF_SEND,
                                         req->comm));
        DB ("rank %d, After _USF_SEND Remove %d\n", rank, req->peer);
      }
      pthread_mutex_unlock (&_USF_mut_sysreq);

      if ((req->peer != MPI_ANY_SOURCE) || sysreq_flag) {
        DB ("rank %d, GO into send recv_post\n", rank, 1);
        /* Send a recv_post */
        hdr.source = rank;
        hdr.dest = req->peer;
        hdr.tag = req->tag;
        hdr.comm = req->comm;
        hdr.size = 0;
        hdr.type = _USF_RECVPOST;
        DB ("rank %d Send recv post\n", rank, 1);
        _USF_send_all (req->peer, sizeof (_USF_message_header), &hdr,
                       "Packet isn't transmitted");
        req->posted = 1;
        recv_post_account[req->peer]++;
      }
    }
    curr = curr->next;
  }
  pthread_mutex_unlock (&_USF_mut_req);
  DB ("rank %d, unlock mut_req before exiting serve_mpi_req\n", rank, 1);
  return;
}

void *
_USF_comm_thread (void *arg)
{
#ifdef _USF_MPI_TCP_
  _USF_TCP_comm_thread ();
#endif
#ifdef _USF_MPI_GM_
  _USF_GM_comm_thread ();
#endif
  pthread_exit ((void *) 1);
}


#endif
