#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <sys/wait.h>
#include <signal.h>
#include "mpi.h"
#include "gm.h"
#include "gm_thread.h"

#ifdef _USF_MPI_GM_
struct gm_port *local_port;     // this is the GM port for communication
int local_port_id;              // this is the GM port ID
int my_board_num = 0;           // the first GM network card number, 
                                // if you have multiple board, 
                                // it should be 0, 1, 2, 3 ...
extern dma_mem_reg_table *mem_reg_table;
void *send_header_list[_USF_MAX_HEADERS_];      // for sending post
void *debug_send_header_list[_USF_MAX_HEADERS_];
int send_header_stack[_USF_MAX_HEADERS_];
int send_header_stack_top = 0;

#define DB(a,b,c)
#ifdef DEBUG
#undef DB
#define DB(a,b,c) printf (a,b,c); fflush (stdout);
#endif /*  */
int _USF_num_of_group = 0;      // group reference num
int _USF_group_max = 64;        // maximum group number
int _USF_group_ref = 0;
int _USF_comm_max = 64;         // maximum communicator number;
int _USF_comm_ref = 200;
struct _USF_rank_type *_USF_comm_list;  /* The communicator list */
struct _USF_group_type *_USF_group_list;        /* The group list */
int _USF_num_of_comm = 0;       /* The number of communicators present */
pthread_t _USF_comm_deamon;     /* The communication thread */
pthread_t _USF_gm_deamon;       /* The gm polling thread thread */
int *_USF_finalize_list;        /* The list that holds the execution completion
                                 * signals */
pthread_mutex_t _USF_mut_sysreq;        /* sys_request link list */
pthread_mutex_t _USF_mut_req;   /* MPI_Request_Int link list */
pthread_mutex_t _USF_mut_block; /* Blocking communication */
pthread_cond_t _USF_cond_block; /* Blocking Communication */
pthread_mutex_t _USF_mut_gm_send;
pthread_cond_t _USF_cond_gm_send;
pthread_mutex_t _USF_mut_gm_thread;
pthread_cond_t _USF_cond_gm_thread;
int _USF_s_pipe[2];             /* 0 read, 1 write */
int _USF_gm_pipe[2];            // GM thread pipe
//int *heads;
#define HEAD_LENGTH sizeof(int)*6

// Used for searching GM id
int _USF_number_proc;

MPI_Request_Int *_USF_synchr_list[_USF_MAXPROCESSORS];

  /* The array that keeps the requests ptrs of ongoing
   * synch communications NULL means no comm. */
void usf_gm_send_callback (struct gm_port *port, void *context,
                           gm_status_t the_status);
int _USF_GM_open_port (struct gm_port **p);
void *_USF_comm_thread (void *arg);
int
_USF_GM_init (int *argc, char ***argv)
{
  int i;
  char *my_name;
  int err;                      // The error value.
  int nump;                     /* Number of processors. */
  int rank;                     /* The rank of the current processor. */
  void *ptr;
  int recv_tokens;

  // initialize GM system
#ifdef DEBUG
  printf (" --> Start to run _USF_GM_init \n");
  fflush (stdout);

#endif

// open GM port
  gm_init ();


  // create communicator
  // create the default comm and group list, the length is 16, 
  // we can enlarge it later when needed
  _USF_comm_list =
    (struct _USF_rank_type *) malloc (_USF_comm_max *
                                      sizeof (struct _USF_rank_type));
  _USF_group_list =
    (struct _USF_group_type *) malloc (_USF_group_max *
                                       sizeof (struct _USF_group_type));
  _USF_comm_list[0].nump = atoi (*((*argv) + 2));
  nump = _USF_comm_list[0].nump;
  _USF_number_proc = nump;
  for (i = 0; i < _USF_group_max; i++) {
    _USF_group_list[i].rank_old = (int *) malloc (nump * sizeof (int));
    _USF_group_list[i].rank_global = (int *) malloc (nump * sizeof (int));
  }

  // Get your own name
  my_name = (char *) malloc (30 * sizeof (char));
  if (gethostname (my_name, 30) != 0)
    return Error (ER_HOSTNM);

  // Assume argc will be true since called internally.
  _USF_comm_list[0].rank = atoi (*((*argv) + 1));
  rank = _USF_comm_list[0].rank;

  // initialize the MPI_COMM_WORLD
  _USF_finalize_list = (int *) malloc (nump * sizeof (int));
  if ((err = _USF_make_conn_list (my_name, nump)) != 0)
    return err;
  _USF_comm_list[0].comm = MPI_COMM_WORLD;
  _USF_comm_list[0].old_comm = MPI_COMM_WORLD;
  _USF_num_of_comm = 1;
  _USF_num_of_group = 1;

  // initialize the group information for MPI_COMM_WORLD
  _USF_comm_list[0].group = 0;
  _USF_group_list[0].group_no_old = 0;
  _USF_group_list[0].group_no_local = 0;
  _USF_group_ref++;             // intial value is 0   
  _USF_group_list[0].group_size = nump;
  for (i = 0; i < nump; i++) {
    _USF_group_list[0].rank_global[i] = _USF_group_list[0].rank_old[i] = i;
  }

  // open a GM port
  _USF_GM_open_port (&local_port);
  gm_allow_remote_memory_access (local_port);

  //provide enough memory to gm_recv
  gm_free_send_tokens (local_port, GM_LOW_PRIORITY,
                       gm_num_send_tokens (local_port));

  // get node_id
  for (i = 0; i < nump; i++) {
    _USF_conn_list[i].node_id =
      gm_host_name_to_node_id (local_port, _USF_conn_list[i].hostname);

  }

  for (i = 0; i < _USF_MAX_HEADERS_; i++) {
    send_header_list[i] =
      gm_dma_malloc (local_port, sizeof (struct _USF_header) * 64);
    debug_send_header_list[i] =
      gm_dma_malloc (local_port, sizeof (struct _USF_header) * 64);
    send_header_stack[i] = i;

  }

  mem_reg_table =
    (dma_mem_reg_table *) malloc (sizeof (dma_mem_reg_table) *
                                  MAX_MEM_REG_ENTRY);
  for (i = 0; i < MAX_MEM_REG_ENTRY; i++) {
    mem_reg_table[i].p = NULL;
    mem_reg_table[i].duplicate_link = NULL;
  }

  // get rid of extra three parameters
  *((*argv) + 3) = *((*argv) + 0);
  (*argv) = (*argv) + 3;
  *argc -= 3;
  return 0;
}

int
_USF_GM_init_comm ()
{
}
int
_USF_GM_open_port (struct gm_port **p)
{
  gm_status_t my_status;
  int i;

  // open local GM port
  local_port_id = _USF_conn_list[_USF_comm_list[0].rank].port_id;
  my_status = gm_open (p, my_board_num, local_port_id, "USFMPI_GM",
                       (enum gm_api_version) GM_API_VERSION_1_1);
  if (my_status != GM_SUCCESS) {
    gm_perror ("[Init] Couldn't open GM port", my_status);
    gm_exit (1);
  }
#ifdef DEBUG
  gm_get_node_id (local_port, &i);
  gm_printf ("[Init] --> Opened board %d port %d on node id %d\n",
             my_board_num, local_port_id, i);

#endif
}

int
_USF_GM_thread_init ()
{

  pthread_mutex_init (&_USF_mut_sysreq, NULL);
  pthread_mutex_init (&_USF_mut_req, NULL);
  pthread_mutex_init (&_USF_mut_block, NULL);
  pthread_cond_init (&_USF_cond_block, NULL);
  gm_thread_init ();

#ifndef SINGLE_THREAD
  if (pthread_create (&_USF_gm_deamon, NULL, gm_thread, NULL) != 0) {
    return Error (ERF_PTHREAD);
  }
#endif /*  */

  //      create GM polling thread here
  return 0;
}

void
_USF_GM_comm_thread ()
{
}
int
_USF_GM_prepare_set (fd_set * set)
{
  int i;                        /* Loop variable */
  int nump;                     /* # of procs in COMM_WORLD */
  int rank;                     /* the rank in COMM_WORLD */
  int max = -1;                 /* The maximum file descriptor */
  MPI_Comm_size (MPI_COMM_WORLD, &nump);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  FD_ZERO (set);
  if (_USF_s_pipe[0] > max)
    max = _USF_s_pipe[0];
  FD_SET (_USF_s_pipe[0], set);

#ifdef DEBUG
  printf("rank--> %i s_pipe[0] %d\n",rank,_USF_s_pipe[0]); fflush(stdout);
#endif /*  */
  if (_USF_gm_pipe[0] > max)
    max = _USF_gm_pipe[0];
  FD_SET (_USF_gm_pipe[0], set);

  return max;
}

int
_USF_GM_send_all (int dest, int size, void *buf, char *err_mes)
{
  int total = 0;
  int send_size;
  int rank;
  int dest_node_id;
  int dest_port_id;
  int i;
  unsigned int test_node_id;
  int *buf_ptr;
  int blocked = 1;
  gm_register_memory (local_port, buf, size);
  dest_node_id = _USF_conn_list[dest].node_id;
  dest_port_id = _USF_conn_list[dest].port_id;

  gm_send_with_callback (local_port, buf, GM_PKT_SIZE,
                         size + 1, GM_LOW_PRIORITY, dest_node_id,
                         dest_port_id, usf_gm_send_callback, &blocked);

  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  pthread_mutex_lock (&_USF_mut_gm_send);
  if (blocked == 1)             //modified by Qing Huang
  {
    pthread_cond_wait (&_USF_cond_gm_send, &_USF_mut_gm_send);
  }
  pthread_mutex_unlock (&_USF_mut_gm_send);

  return size;
}

int
_USF_GM_Finalize ()
{
  int i;
  dma_mem_reg_table *pre, *next;
  for (i = 0; i < _USF_group_max; i++) {
    free (_USF_group_list[i].rank_old);
    free (_USF_group_list[i].rank_global);
  }

#ifdef DEBUG
  printf ("Start to free header buffer\n");
  fflush (stdout);

#endif
  for (i = 0; i < _USF_MAX_HEADERS_; i++) {
    gm_dma_free (local_port, send_header_list[i]);
  }

#ifdef DEBUG
  printf ("Start to deregister DMA buffer\n");
  fflush (stdout);
#endif
  for (i = 0; i < MAX_MEM_REG_ENTRY; i++) {
    if (mem_reg_table[i].p != NULL) {
      gm_deregister_memory (local_port, mem_reg_table[i].p,
                            mem_reg_table[i].size);
      next = mem_reg_table[i].duplicate_link;
      while (next != NULL) {
        pre = next;
        next = next->duplicate_link;
        gm_deregister_memory (local_port, pre->p, pre->size);
        free (pre);
      }
    }
  }

#ifdef DEBUG
  printf ("Start to free local memory data\n");
  fflush (stdout);

#endif
  free (mem_reg_table);
  free (_USF_comm_list);        // use array, not list
  free (_USF_group_list);
  free (_USF_conn_list);
  free (mem_reg_table);

#ifdef DEBUG
  printf ("Start to close GM\n");
  fflush (stdout);

#endif /*  */

#ifdef DEBUG
  printf ("Start to GM finalize\n");
  fflush (stdout);

#endif /*  */
  gm_finalize ();

#ifdef DEBUG
  printf ("Finish all GM\n");
  fflush (stdout);

#endif /*  */
}

void
usf_gm_send_callback (struct gm_port *port, void *context,
                      gm_status_t the_status)
{

  // One pending callback has been received
  //
  int rank;
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);

  pthread_mutex_lock (&_USF_mut_gm_send);
  (*(int *) context)--;
  pthread_mutex_unlock (&_USF_mut_gm_send);
  pthread_cond_signal (&_USF_cond_gm_send);
  if (the_status != GM_SUCCESS) {
    printf ("GM sending is wrong\n");
    fflush (stdout);
    if (the_status != GM_SEND_DROPPED) {
      gm_perror ("send completed with error\n", the_status);
      exit (1);
    }
  }
}


//-----------------------------------------------------------------------
// Added by Chaney
// get_rank_by_gm_id : Translate gm id to rank.
// Parameters :
// id : GM id.
// rank : mpi rank
//-----------------------------------------------------------------------
void
get_rank_by_gm_id (int node_id, int port_id, int *rank)
{
  int i;
  for (i = 0; i < _USF_number_proc; i++) {
    if (_USF_conn_list[i].node_id == node_id
        && _USF_conn_list[i].port_id == port_id) {
      *rank = i;
      return;
    }
  }
  printf ("Can't find gm id %d.", node_id);
  fflush (stdout);
  *rank = -100;
}


#endif
