//-------------------------------------------------------------------------
// Gm_thread.c : GM message polling thread. Used for listening incomming 
// message
// Author : Chaney Tsai 9/12/2002
// University of San Francisco
//-------------------------------------------------------------------------
#include <string.h>
#include <stdio.h>
#include "mpi.h"
#include "gm.h"
#include "gm_thread.h"

//#define DEBUG

#ifdef _USF_MPI_GM_
//#define USF_GM_DEBUG

#define DB(a,b,c)
#ifdef DEBUG
#undef DB
#define DB(a,b,c) printf (a,b,c); fflush (stdout);
#endif /*  */
int thread_recving;
int gm_thread_polling = 1;

//gm data queue node
typedef struct gm_msg_node
{
  int size;
  void *buf;
  int next;
  int prev;
}
gm_msg_node;
typedef struct gm_conn_node
{
  int head;
  int tail;
}
gm_conn_node;
gm_conn_node *gm_conn_msg_queue;
gm_msg_node *gm_msg_queue;
int gm_msg_queue_head;
pthread_mutex_t _USF_mut_gm_recv;
pthread_mutex_t _USF_mut_gm_pending;    // added by Qing
int request_pending = 0;
pthread_cond_t _USF_cond_gm_recv;
int waiting_for_data;
static int my_rank;
static int total_nodes;
int blocking_recv;
int ack_pending = 0;
extern void *send_header_list[_USF_MAX_HEADERS_];       // for sending post
extern void *debug_send_header_list[_USF_MAX_HEADERS_];
extern int send_header_stack[_USF_MAX_HEADERS_];
extern int send_header_stack_top;
void *request_list[_USF_MAX_HEADERS_];
int send_request_counter = 0;
int recv_request_counter = 0;
int message_counter = 0;
int recv_post_counter = 0;
int ack_post_counter = 0;
int send_post_counter = 0;
int direct_mem_counter = 0;
int direct_mem_callback_counter = 0;
int dest_node_flag = -1;
int dest_ack_flag = -1;
extern int isend_call;
extern int recv_call;
int ack_signal_counter = 0;
int ack_request_counter = 0;
int debug_counter = 0;
MPI_Request_Int *debug_current_req = NULL;
MPI_Request_Int *debug_current_req0 = NULL;

// gm_thread_int : Thread end function
// Set thread_recving to end gm polling thread.
void
gm_thread_init (void)
{
  void *ptr;
  int recv_tokens;
  int i;
  MPI_Comm_size (MPI_COMM_WORLD, &total_nodes);

  //Initial connection message queue
  //Provide buffer pool to GM
  recv_tokens =
    ((gm_num_receive_tokens (local_port) >
      GM_MAX_RECV_TOKENS) ? GM_MAX_RECV_TOKENS :
     gm_num_receive_tokens (local_port));
  for (i = 0; i < recv_tokens * 2; i++) {
    ptr = gm_dma_malloc (local_port, _USF_MAXDATASIZE);
    gm_provide_receive_buffer (local_port, ptr, _USF_MAXDATASIZE,
                               GM_LOW_PRIORITY);

#ifdef USF_GM_DEBUG
    printf ("Addredd = %p.\n", ptr);
    fflush (stdout);
#endif /*  */
  }
  MPI_Comm_rank (MPI_COMM_WORLD, &my_rank);
}


// gm_thread_end : Thread end function
// Set thread_recving to end gm polling thread.
void
gm_thread_end (void)
{
  thread_recving = 0;
}


int
_USF_send_header_pop ()
{

  if (send_header_stack_top < _USF_MAX_HEADERS_) {
    send_header_stack_top++;
    return send_header_stack[send_header_stack_top - 1];
  }

  else {
    printf ("send stack pop error\n");
    fflush (stdout);
    return -1;
  }
}
void
_USF_send_header_push (int free_header)
{
  if (send_header_stack_top > 0) {
    send_header_stack_top--;
    send_header_stack[send_header_stack_top] = free_header;
  }

  else
    printf ("rank %d error of push op\n", my_rank);
}

void
gm_send_post_callback (struct gm_port *port, void *context,
                       gm_status_t the_status)
{
  struct _USF_header *header, *debug_header;
  _USF_send_header_push ((int) context);

#ifdef SINGLE_THREAD
  thread_recving = 0;

#endif /*  */

}

void
_USF_gm_send_post (struct _USF_header *header, int i)
{
  gm_send_with_callback (local_port, header, _USF_MAXDATASIZE,
                         sizeof (struct _USF_header), GM_LOW_PRIORITY,
                         _USF_conn_list[header->dest].node_id,
                         _USF_conn_list[header->dest].port_id,
                         gm_send_post_callback, (void *) i);
}

void
gm_send_ack_post_callback (struct gm_port *port, void *context,
                           gm_status_t the_status)
{

  struct _USF_header *header, *debug_header;

  _USF_send_header_push ((int) context);
  ack_pending--;

#ifdef SINGLE_THREAD
  thread_recving = 0;

#endif /*  */

#ifdef USF_GM_DEBUG
  printf ("rank %d ark is over\n", my_rank);
  fflush (stdout);

#endif /*  */
  ack_request_counter++;

}

void
gm_handle_recv_post_callback (struct gm_port *port, void *context,
                              gm_status_t the_status)
{
  int i;
  struct _USF_header *header, *debug_header;
  if (the_status != GM_SUCCESS) {
    printf ("rank %d send data error\n", my_rank);
    fflush (stdout);
  }
  i = _USF_send_header_pop ();
  if (i == -1) {
    printf ("running out of header\n");
    fflush (stdout);
    exit (1);
  }
  header = (struct _USF_header *) send_header_list[i];
  header->source = my_rank;

  header->dest = ((MPI_Request_Int *) context)->peer;
  header->tag = ((MPI_Request_Int *) context)->tag;

  header->comm = ((MPI_Request_Int *) context)->comm;
  header->type = _USF_ACK_GMSEND;
  header->size = 0;

#ifdef USF_GM_DEBUG
  printf ("rank %d send ACK header to rank %d\n", my_rank, header->dest);
  fflush (stdout);

#endif
  ack_pending++;

  gm_send_with_callback (local_port, header, _USF_MAXDATASIZE,
                         sizeof (struct _USF_header), GM_LOW_PRIORITY,
                         _USF_conn_list[header->dest].node_id,
                         _USF_conn_list[header->dest].port_id,
                         gm_send_ack_post_callback, (void *) i);

  dest_ack_flag = _USF_conn_list[header->dest].node_id;

  pthread_mutex_lock (&_USF_mut_block);
  if (((MPI_Request_Int *) context)->blocking) {
    ((MPI_Request_Int *) context)->end_of_transmission = 1;
    ((MPI_Request_Int *) context)->done = 1;
    pthread_cond_signal (&_USF_cond_block);
  }

  else {
    ((MPI_Request_Int *) context)->done = 1;
    ((MPI_Request_Int *) context)->end_of_transmission = 1;
  }
  pthread_mutex_unlock (&_USF_mut_block);

}

void
_USF_gm_handle_recv_post (int dest, int source, int tag, int comm,
                          int in_size, void *in_buf, void *out_buf,
                          int out_size, MPI_Request_Int * req)
{
  int actual_size;
  int dest_node_id;
  int dest_port_id;
  gm_status_t temp;
  actual_size = (in_size <= out_size) ? in_size : out_size;
  dest_node_id = _USF_conn_list[dest].node_id;
  dest_node_flag = dest_node_id;
  dest_port_id = _USF_conn_list[dest].port_id;

  if (actual_size == 0) {
    temp = GM_SUCCESS;

    gm_handle_recv_post_callback (local_port, req, temp);
    return;
  }
  DB ("rank %d strat to send data\n", my_rank, 1)
    gm_directed_send_with_callback (local_port, out_buf,
                                    (gm_remote_ptr_t) (gm_up_t) in_buf,
                                    actual_size, GM_LOW_PRIORITY,
                                    dest_node_id, dest_port_id,
                                    gm_handle_recv_post_callback, req);
}

void
gm_handle_header (struct _USF_header *h_ptr)
{

  int i;                        /* Loop variable */
  int nump;                     /* # of procs in COMM_WORLD */

  int recv_size;                /* The size of received packet */
  int size;                     /* Used for ready mode receive */
  _USF_request_node *req1, *req_probe;  /* The request node used for lookup */
  _USF_sysreq_node *sysreq1;    /* The sysreq node used to insert a node */
  int isongcomm = 0;            /* Is there an ongoiing communication */
  int header_msg = 0;
  int source;
  struct _USF_header *header;
  MPI_Request_Int *req;
  _USF_request_node *node = _USF_request_hd;
  source = h_ptr->source;
  switch (h_ptr->type) {
  case _USF_RECVPOST:
    recv_post_counter++;

    /* These are removed from sysreq by MPI_(I)SEND calls */
    /* Look for a send request in request list */
    DB ("rank %d, Get a recv_post\n", my_rank, 1);
    pthread_mutex_lock (&_USF_mut_req);
    req1 =
      _USF_request_lookup (h_ptr->tag, h_ptr->source, _USF_SEND, h_ptr->comm);
    DB ("rank %d, isong %d\n", my_rank, isongcomm);
    if (req1 == NULL) {
      sysreq1 = (_USF_sysreq_node *) pool_take (_USF_sysreq_node_pool);
      if (sysreq1 == NULL)
        DB ("rank %d Memory error", my_rank, 1);
      pthread_mutex_lock (&_USF_mut_sysreq);
      _USF_sysreq_insert (h_ptr->tag, h_ptr->source, _USF_RECEIVE,
                          h_ptr->comm, h_ptr->size, h_ptr->in_buf, sysreq1);
      pthread_mutex_unlock (&_USF_mut_sysreq);
    } else {
      DB ("rank %d, handle recv_post\n", my_rank, 1);
      _USF_gm_handle_recv_post (h_ptr->source, h_ptr->dest, h_ptr->tag,
                                h_ptr->comm, h_ptr->size, h_ptr->in_buf,
                                req1->req->buf, req1->req->size, req1->req);

      /* update mpi_request */
      req1->req->done = 1;
      if (req1->req->posted == 0)
        ++send_post_account[h_ptr->source];

      /* Check if it was blocking !!! */
    }
    pthread_mutex_unlock (&_USF_mut_req);
    break;
  case _USF_SENDPOST:
    send_post_counter++;

    /* These are removed from sysreq by MPI_(I)RECV calls */
    /* Send a recv post if unsent recv post in mpi req list */
    DB ("rank %d Get a send_post\n", my_rank, 1);
    if (h_ptr->send_seq_num <= recv_post_account[h_ptr->source])
      break;
    pthread_mutex_lock (&_USF_mut_req);

    /* Continue with the operation */
    req1 = _USF_request_lookup_p (h_ptr->tag, _USF_RECEIVE, h_ptr->comm, 0);
    if (req1 == NULL) {         /* put in the sys req queue */

      //Received a send-post but there is no recv req in system queue
      DB ("rank %d SENDPOST NULL, put in sysreq Q\n", my_rank, 1);
      sysreq1 = (_USF_sysreq_node *) pool_take (_USF_sysreq_node_pool);
      pthread_mutex_lock (&_USF_mut_sysreq);
      _USF_sysreq_insert (h_ptr->tag, h_ptr->source, _USF_SEND, h_ptr->comm,
                          h_ptr->size, h_ptr->in_buf, sysreq1);
      pthread_mutex_unlock (&_USF_mut_sysreq);

      /* check if there is a probe */
      req_probe =
        _USF_request_lookup_p (MPI_ANY_TAG, _USF_PROBE, MPI_COMM_WORLD, 0);
      if (req_probe != NULL) {

        pthread_mutex_lock (&_USF_mut_block);
        if (req_probe->req->blocking) {
          req_probe->req->end_of_transmission = 1;
          pthread_cond_signal (&_USF_cond_block);
        }

        else
          req_probe->req->end_of_transmission = 1;
        pthread_mutex_unlock (&_USF_mut_block);
        thread_recving = 0;
      }
    } else {

      /* will match a MPI_ANY_SOURCE nonsent recv */
      /* Send a recv_post */
      if (req1->req->peer != MPI_ANY_SOURCE) {
        pthread_mutex_unlock (&_USF_mut_req);
        DB ("node %d don't do anything for send header\n", my_rank, 1);
        break;
      }
      DB ("node %d send recv post for any source to node %d\n", my_rank,
          h_ptr->source);
      req = req1->req;
      i = _USF_send_header_pop ();
      if (i == -1) {
        printf ("running out of header\n");
        fflush (stdout);
        exit (1);
      }
      header = (struct _USF_header *) send_header_list[i];
      header->source = my_rank;
      header->dest = h_ptr->source;
      req->peer = h_ptr->source;
      header->tag = req->tag;
      header->comm = req->comm;
      header->type = _USF_RECVPOST;
      header->size = req->size;
      header->in_buf = req->buf;
      _USF_gm_send_post (header, i);

      req1->req->posted = 1;
      recv_post_account[req1->req->peer]++;
    }
    pthread_mutex_unlock (&_USF_mut_req);
    break;
  case _USF_ACK_GMSEND:
    pthread_mutex_lock (&_USF_mut_req);
    DB ("rank %d, get a ack post, done receiving!\n", my_rank, 1);
    req1 =
      _USF_request_lookup_m (h_ptr->tag, h_ptr->source, _USF_RECEIVE,
                             h_ptr->comm, 0);
    ack_post_counter++;
    if (req1 != NULL) {
      pthread_mutex_lock (&_USF_mut_block);
      ack_signal_counter++;
      req1->req->done = 1;

      /* Check if it was blocking !!! */
      if (req1->req->blocking) {
        DB ("rank %d, wake up main\n", my_rank, 1);
        req1->req->end_of_transmission = 1;
        pthread_cond_signal (&_USF_cond_block);
      }

      else
        req1->req->end_of_transmission = 1;

#ifdef SINGLE_THREAD
      thread_recving = 0;

#endif
      pthread_mutex_unlock (&_USF_mut_block);
    } else {
      DB ("rank %d can't find RECV request, error ACK header\n", my_rank, 1);
      printf
        ("rank %d bad header source %d, dest %d, type %d comm %d, tag %d, size %d, \n",
         my_rank, h_ptr->source, h_ptr->dest, h_ptr->type, h_ptr->comm,
         h_ptr->tag, h_ptr->size);
      fflush (stdout);

      while (node != NULL) {
        node = node->next;
      }
    }
    pthread_mutex_unlock (&_USF_mut_req);
    break;
  default:
    DB ("rank %d , error header\n", my_rank, 1);
    break;
  }
}
void
_USF_gm_serve_mpi_requests (int rank, int nump)
{
  int i;                        /* Loop variable */
  int args[7];                  /* The message header */
  _USF_request_node *curr;      /* The pointer used to traverse the list */
  MPI_Request_Int *req;         /* The pointer used to store MPI_Request_Int */
  _USF_sysreq_node *sysreq1;    /* The pointer used to retrieve sysreq */
  int send_size;                /* The number of bytes sent */
  int sysreq_flag;              /* Shows if there was a matching sysreq */
  struct _USF_header *header;
  pthread_mutex_lock (&_USF_mut_req);
  curr = _USF_request_hd;

  /* Traverse and serve all the requests */
  while (curr != NULL) {
    req = curr->req;
    DB ("rank %d s_mpi_req fst type %d\n", rank, req->type);

    /* continue if already completed */
    if ((req->done) || (req->posted) || (req->active == 0)) {
      DB ("rank %d found completed or posted or inactive\n", rank, 1);
      curr = curr->next;
      continue;
    }
    if (req->type == _USF_SEND) {       /* type send */
      send_request_counter++;
      pthread_mutex_lock (&_USF_mut_sysreq);

      /* Check for a matching sysreq recv post */
      sysreq1 =
        _USF_sysreq_lookup (req->tag, req->peer, _USF_RECEIVE, req->comm);
      DB ("rank %d serve mpi _USF_SEND request.\n", rank, 1);
      if (sysreq1 == NULL) {    /* send send_post */

        /* if it is a ready mode message send
         * immediately */
        DB ("rank %d Send a send_post\n", rank, 1);
        i = _USF_send_header_pop ();
        if (i == -1) {
          printf ("running out of header\n");
          fflush (stdout);
          exit (1);
        }
        header = (struct _USF_header *) send_header_list[i];
        header->source = my_rank;
        header->dest = req->peer;
        header->tag = req->tag;
        header->comm = req->comm;
        header->type = _USF_SENDPOST;
        header->size = req->size;
        header->in_buf = NULL;
        header->send_seq_num = ++send_post_account[req->peer];
        _USF_gm_send_post (header, i);
        req->posted = 1;
        DB ("rank %d Sent a send_post!!!\n", rank, 1);
      } else {                  /* if there is a request send the message */
        send_post_account[req->peer]++;
        DB ("rank %d serve recv_post in MPI_serv request.\n", rank, 1);
        _USF_gm_handle_recv_post (sysreq1->source, rank, sysreq1->tag,
                                  sysreq1->comm, sysreq1->size,
                                  sysreq1->in_buf, req->buf, req->size, req);
        req->done = 1;
        pool_return (_USF_sysreq_node_pool, (pool_item *)
                     _USF_sysreq_remove (req->tag, req->peer, _USF_RECEIVE,
                                         req->comm));
      }
      pthread_mutex_unlock (&_USF_mut_sysreq);
    } else if (req->type == _USF_RECEIVE) {     /* type receive */
      recv_request_counter++;
      DB ("rank %d serve _USF_RECEIVE request\n", rank, 1);
      pthread_mutex_lock (&_USF_mut_sysreq);

      /* Check for a matching sysreq send post */
      DB ("rank %d, req->peer %d\n", rank, req->peer);
      sysreq1 =
        _USF_sysreq_lookup (req->tag, req->peer, _USF_SEND, req->comm);

      /* If there is remove it */
      if (sysreq1 == NULL) {
        DB ("rank %d, NOT FOUND SYSREQ\n", rank, 1);
        sysreq_flag = 0;
      } else {
        DB ("rank %d, FOUND SYSREQ src %d\n", rank, sysreq1->source);
        sysreq_flag = 1;
        req->peer = sysreq1->source;
        pool_return (_USF_sysreq_node_pool, (pool_item *)
                     _USF_sysreq_remove (req->tag, req->peer, _USF_SEND,
                                         req->comm));
        DB ("rank %d, After _USF_SEND Remove %d\n", rank, req->peer);
      }
      pthread_mutex_unlock (&_USF_mut_sysreq);
      if ((req->peer != MPI_ANY_SOURCE) || sysreq_flag) {
        DB ("rank %d, GO into send recv_post\n", rank, 1);
        i = _USF_send_header_pop ();
        if (i == -1) {
          printf ("running out of header\n");
          fflush (stdout);
          exit (1);
        }
        header = (struct _USF_header *) send_header_list[i];
        header->source = my_rank;
        header->dest = req->peer;
        header->tag = req->tag;
        header->comm = req->comm;
        header->type = _USF_RECVPOST;
        header->size = req->size;
        header->in_buf = req->buf;
        _USF_gm_send_post (header, i);
        DB ("rank %d, sent recv_post\n", rank, 1);
        req->posted = 1;
        recv_post_account[req->peer]++;
      }
    }
    curr = curr->next;
  }
  pthread_mutex_unlock (&_USF_mut_req);
  DB ("rank %d, unlock mut_req before exiting serve_mpi_req\n", my_rank, 1);
  return;
}


// gm_thread : Thread run function
// This is the body of listening thread function. This function only polling
// (receiving) 6-int headers/POSTs
void *
gm_thread ()
{
  gm_recv_event_t *event;
  int sender_rank = 0;
  int fast_recv;
  struct _USF_header *ptr, pre_header;
  int temp_counter;

  thread_recving = 1;
  blocking_recv = 0;

  //Waiting for incomming message
  while (thread_recving) {
    debug_counter++;
    //handle local recv request
    while (request_pending) {
      _USF_gm_serve_mpi_requests (my_rank, total_nodes);
      pthread_mutex_lock (&_USF_mut_gm_pending);
      request_pending--;
      pthread_mutex_unlock (&_USF_mut_gm_pending);
    }

    event = gm_receive (local_port);

    switch (gm_ntoh_u8 (event->recv.type)) {
    case GM_NO_RECV_EVENT:
      break;
    case GM_FAST_RECV_EVENT:
    case GM_FAST_PEER_RECV_EVENT:
    case GM_FAST_HIGH_RECV_EVENT:
    case GM_FAST_HIGH_PEER_RECV_EVENT:
      DB ("node %d gets gm header\n", my_rank, 1);

      ptr = (struct _USF_header *) (gm_ntohp (event->recv.message));

      gm_handle_header (ptr);

      gm_provide_receive_buffer (local_port, gm_ntohp (event->recv.buffer),
                                 _USF_MAXDATASIZE, GM_LOW_PRIORITY);
      break;

    case GM_RECV_EVENT:
    case GM_PEER_RECV_EVENT:
    case GM_HIGH_RECV_EVENT:
    case GM_HIGH_PEER_RECV_EVENT:

      //Handle incoming mssage
      printf ("message error\n");
      fflush (stdout);
      break;
    default:
      gm_unknown (local_port, event);
    }
  }
}


//-------------------------------------------------------------------------
// _USF_GM_recv_all : Receive incomming messages
// This has the same functionality as _USF_TCP_recv_all. It receives data 
// from GM and determine if this is the data that user looks for.
// If not, it queues the data and do receive again until it receive the 
// correct data.
//-------------------------------------------------------------------------
void
_USF_GM_recv_all (int source, int size, void *buf, char *msg)
{
}
#endif 
