#ifndef _MPI_H
#define _MPI_H 1

#include <pthread.h>
#include "pool.h"

#define _USF_SERVPORT 10239
#define _USF_MAXDATASIZE 1024*64  
#define _USF_MAXPROCESSORS 128

// - Begin - by Qing Huang

#define SINGLE_THREAD
//#define DEBUG
#define _USF_MPI_TCP_
//#define _USF_MPI_GM_


/* If TCP is used  - Macros are used for now, function indirection may
   be used later */


#ifdef _USF_MPI_TCP_
#define _USF_send_all 		_USF_TCP_send_all
#define _USF_recv_all 		_USF_TCP_recv_all
#define _USF_finalize 		_USF_TCP_Finalize
#endif

#ifdef _USF_MPI_GM_
#define _USF_send_all 		_USF_GM_send_all
#define _USF_recv_all 		_USF_GM_recv_all
#define _USF_finalize		_USF_GM_Finalize
#define _USF_MAX_HEADERS_ 64
#define _USF_MAX_DMA_DATA_BUFFER 16
#define MAX_MEM_REG_ENTRY 1024

void *data_buffer[_USF_MAX_DMA_DATA_BUFFER];


typedef struct dma_mem_reg_table
{
  void *p;
  int size;
  struct dma_mem_reg_table *duplicate_link;
}
dma_mem_reg_table;


#endif

unsigned int send_post_account[_USF_MAXPROCESSORS];
unsigned int recv_post_account[_USF_MAXPROCESSORS];

//unsigned int sending_requests[_USF_MAXPROCESSORS] ;
//unsigned int recv_pending[_USF_MAXPROCESSORS] ;


/* added by Qing Huang */

#define MPI_COMM_NULL      ((MPI_Comm)0)
#define MPI_OP_NULL        ((MPI_Op)0)
#define MPI_GROUP_NULL     ((MPI_Group)0)
#define MPI_DATATYPE_NULL  ((MPI_Datatype)0)
#define MPI_REQUEST_NULL   ((MPI_Request)0)
#define MPI_ERRHANDLER_NULL ((MPI_Errhandler )0)
typedef int MPI_Errhandler;
typedef int MPI_Group;
/* Pre-defined constants */
#define MPI_UNDEFINED      (-32766)
#define MPI_UNDEFINED_RANK MPI_UNDEFINED
#define MPI_KEYVAL_INVALID 0

#define MPI_SUCCESS          0  /* Successful return code */
/* Communication argument parameters */
#define MPI_ERR_BUFFER       1  /* Invalid buffer pointer */
#define MPI_ERR_COUNT        2  /* Invalid count argument */
#define MPI_ERR_TYPE         3  /* Invalid datatype argument */
#define MPI_ERR_TAG          4  /* Invalid tag argument */
#define MPI_ERR_COMM         5  /* Invalid communicator */
#define MPI_ERR_RANK         6  /* Invalid rank */
#define MPI_ERR_ROOT         7  /* Invalid root */
#define MPI_ERR_TRUNCATE    14  /* Message truncated on receive */

/* MPI Objects (other than COMM) */
#define MPI_ERR_GROUP        8  /* Invalid group */
#define MPI_ERR_OP           9  /* Invalid operation */
#define MPI_ERR_REQUEST     19  /* Invalid mpi_request handle */

/* Special topology argument parameters */
#define MPI_ERR_TOPOLOGY    10  /* Invalid topology */
#define MPI_ERR_DIMS        11  /* Invalid dimension argument */

/* All other arguments.  This is a class with many kinds */
#define MPI_ERR_ARG         12  /* Invalid argument */

/* Other errors that are not simply an invalid argument */
#define MPI_ERR_OTHER       15  /* Other error; use Error_string */

#define MPI_ERR_UNKNOWN     13  /* Unknown error */
#define MPI_ERR_INTERN      16  /* Internal error code    */

/* Multiple completion has two special error classes */
#define MPI_ERR_IN_STATUS   17  /* Look in status for error value */
#define MPI_ERR_PENDING     18  /* Pending request */

/* New MPI-2 Error classes */
#define MPI_ERR_FILE        27  /* */
#define MPI_ERR_ACCESS      20  /* */
#define MPI_ERR_AMODE       21  /* */
#define MPI_ERR_BAD_FILE    22  /* */
#define MPI_ERR_FILE_EXISTS 25  /* */
#define MPI_ERR_FILE_IN_USE 26  /* */
#define MPI_ERR_NO_SPACE    36  /* */
#define MPI_ERR_NO_SUCH_FILE 37 /* */
#define MPI_ERR_IO          32  /* */
#define MPI_ERR_READ_ONLY   40  /* */
#define MPI_ERR_CONVERSION  23  /* */
#define MPI_ERR_DUP_DATAREP 24  /* */
#define MPI_ERR_UNSUPPORTED_DATAREP   43        /* */

/* MPI_ERR_INFO is NOT defined in the MPI-2 standard.  I believe that
   this is an oversight */
#define MPI_ERR_INFO        28  /* */
#define MPI_ERR_INFO_KEY    29  /* */
#define MPI_ERR_INFO_VALUE  30  /* */
#define MPI_ERR_INFO_NOKEY  31  /* */

#define MPI_ERR_NAME        33  /* */
#define MPI_ERR_NOMEM       34  /* */
#define MPI_ERR_NOT_SAME    35  /* */
#define MPI_ERR_PORT        38  /* */
#define MPI_ERR_QUOTA       39  /* */
#define MPI_ERR_SERVICE     41  /* */
#define MPI_ERR_SPAWN       42  /* */
#define MPI_ERR_UNSUPPORTED_OPERATION 44        /* */
#define MPI_ERR_WIN         45  /* */

#define MPI_ERR_LASTCODE    0x3FFFFFFF  /* Last error code */


/* These are only guesses; make sure you change them in mpif.h as well */
#define MPI_MAX_PROCESSOR_NAME 256
#define MPI_MAX_ERROR_STRING   512
#define MPI_MAX_NAME_STRING     63      /* How long a name do you need ? */

typedef long MPI_Aint;

/* Upper bound on the overhead in bsend for each message buffer */
#define MPI_BSEND_OVERHEAD 512

/* Datatypes */
typedef int MPI_Datatype;

//--- Add by Chaney, 6/17/02
//---
#define MPI_BOTTOM      (void *)0
#define MPI_SUCCESS          0  /* Successful return code */
#define MPI_Primitive_Type   64

typedef struct _USF_MPI_struct_type
{
  int idx;
  int count;
  int *block_lengths;
  char *buf;
  MPI_Aint *disp;
  MPI_Datatype *typelist;
  struct _USF_MPI_struct_type *next;
}
_USF_MPI_struct_type;

// User define function
typedef void (MPI_User_function) (void *, void *, int *, MPI_Datatype *);
//---

#define MPI_CHAR           ((MPI_Datatype)1)
#define MPI_SHORT          ((MPI_Datatype)2)
#define MPI_INT            ((MPI_Datatype)3)
#define MPI_LONG           ((MPI_Datatype)4)
#define MPI_UNSIGNED_CHAR  ((MPI_Datatype)5)
#define MPI_UNSIGNED_SHORT ((MPI_Datatype)6)
#define MPI_UNSIGNED       ((MPI_Datatype)7)
#define MPI_UNSIGNED_LONG  ((MPI_Datatype)8)
#define MPI_FLOAT          ((MPI_Datatype)9)
#define MPI_DOUBLE         ((MPI_Datatype)10)
#define MPI_LONG_DOUBLE    ((MPI_Datatype)11)
#define MPI_BYTE           ((MPI_Datatype)12)
#define MPI_LONG_LONG      ((MPI_Datatype)13)

#define MPI_PACKED         ((MPI_Datatype)14)
#define MPI_LB             ((MPI_Datatype)15)
#define MPI_UB             ((MPI_Datatype)16)

/*
   The layouts for the types MPI_DOUBLE_INT etc are simply
   struct {
       double var;
       int    loc;
   }
*/
#define MPI_FLOAT_INT         ((MPI_Datatype)17)
#define MPI_DOUBLE_INT        ((MPI_Datatype)18)
#define MPI_LONG_INT          ((MPI_Datatype)19)
#define MPI_SHORT_INT         ((MPI_Datatype)20)
#define MPI_2INT              ((MPI_Datatype)21)
#define MPI_LONG_DOUBLE_INT   ((MPI_Datatype)22)

/* Special Structs for MINLOC, MAXLOC operations */

typedef struct _USF_FLOAT_INT
{
  float fst;
  int sec;
}
_USF_FLOAT_INT;

typedef struct _USF_DOUBLE_INT
{
  double fst;
  int sec;
}
_USF_DOUBLE_INT;

typedef struct _USF_LONG_INT
{
  long fst;
  int sec;
}
_USF_LONG_INT;

typedef struct _USF_SHORT_INT
{
  short fst;
  int sec;
}
_USF_SHORT_INT;

typedef struct _USF_2INT
{
  int fst;
  int sec;
}
_USF_2INT;

typedef struct _USF_LONG_DOUBLE_INT
{
  long double fst;
  int sec;
}
_USF_LONG_DOUBLE_INT;


/* Collective operations */
typedef int MPI_Op;

#define MPI_MAX    (MPI_Op)(100)
#define MPI_MIN    (MPI_Op)(101)
#define MPI_SUM    (MPI_Op)(102)
#define MPI_PROD   (MPI_Op)(103)
#define MPI_LAND   (MPI_Op)(104)
#define MPI_BAND   (MPI_Op)(105)
#define MPI_LOR    (MPI_Op)(106)
#define MPI_BOR    (MPI_Op)(107)
#define MPI_LXOR   (MPI_Op)(108)
#define MPI_BXOR   (MPI_Op)(109)
#define MPI_MINLOC (MPI_Op)(110)
#define MPI_MAXLOC (MPI_Op)(111)




typedef struct MPI_Status
{
  int count;
  int MPI_SOURCE;
  int MPI_TAG;
  int MPI_ERROR;
}
MPI_Status;

typedef struct MPI_Request_Int
{
  int active;                   /* 0 if not active, 1 if active, 2 if persistent act */
  int done;                     /* Operation is complete */
  int type;                     /* 0 recv, 1 send, 2 probe */
  int blocking;                 /* 0 non-blocking, 1 blocking */
  int posted;                   /* 0 post not sent, 1 post sent */
  void *buf;                    /* The message buffer */
  int size;                     /* The size of the buffer */
  int tag;                      /* The tag of the message */
  int peer;                     /* The other peer of communication */
  int comm;                     /* The communicator of the message */
  int multiple;                 /* 1 if multiple chunks, 0 if one chunk */
  int nchunk;                   /* Number of chunks sent or received */
  //Add by Chaney 2002, used for derives datatype */
  //---
  _USF_MPI_struct_type *stype;  /* Pointer to derived datatype */
  int dcount;                   /* Counter used for send/receive */
  //---
  struct _USF_request_node *node;       /* Back pointer for easy removal */
  int end_of_transmission;      /* to sychronize command thread and main thread add by Qing Huang */
}
MPI_Request_Int;

typedef MPI_Request_Int *MPI_Request;

typedef struct _USF_request_node
{
  struct _USF_request_node *next;       /* required for the link list */
  struct _USF_request_node *back;       /* required for the link list */
  MPI_Request_Int *req;
}
_USF_request_node;

typedef struct _USF_sysreq_node
{
  struct _USF_sysreq_node *next;        /* required for hash table */
  int tag;                      /* the tag of the message */
  int source;                   /* the source of the message */
  int type;                     /* 0 recv request, 1 send request */
  int comm;                     /* The communicator */
  int size;                     /* The size of the message */
#ifdef _USF_MPI_GM_
  void *in_buf;                 // for remote direct memory access, added by Qing Huang
#endif
}
_USF_sysreq_node;




#define _USF_HEADER_TYPE 0

#define _USF_UNIX_SOC 0
#define _USF_TCP_SOC  1

/* Communicators */
typedef int MPI_Comm;
#define MPI_COMM_WORLD 91
#define MPI_COMM_SELF  92

#define MPI_ANY_SOURCE  (-2)
#define MPI_ROOT        (-3)
#define MPI_ANY_TAG     (-1)

//#define MPI_REQUEST_NULL NULL

#define ER_INCORRECT   1
#define ER_FILENAME    2
#define ER_NUMP        3
#define ER_NEMACS      4
#define ER_HOSTNM      5
#define ER_MACFILE     6
#define ERF_SOCKETINIT 30
#define ERF_BINDINIT   31
#define ERF_ACCEPT     32
#define ERF_RECV       33
#define ERF_SEND       34
#define ERF_GETHE      35
#define ERF_CONNECT    36
#define ERF_SETSOCKOPT 37
#define ERF_LISTEN     38
#define ERF_SELECT     39
#define ERF_PTHREAD    40

/* Message Packet
   args header
   0 - source rank
   1 - destination rank
   2 - tag
   3 - communicator
   4 - size
   5 - type:
     0) Recv post
     1) Send post
     2) First chunk of synch communication
     3) Non terminal, non first chunk of synch communication
     4) Terminal chunk of synch communication
     5) Acknowledgement of chunk     
     6) Short message
     7) Communication termination message
     
*/

/*
#define _USF_SOURCE 0
#define _USF_DEST   1
#define _USF_TAG    2
#define _USF_COMM   3
#define _USF_SIZE   4
#define _USF_TYPE   5
*/

typedef struct _USF_message_header
{
  int source;
  int dest;
  int tag;
  int comm;
  int size;
  unsigned int send_seq_num;
  int type;
}
_USF_message_header;

typedef struct _USF_conn_struct
{
  int sd;                       /* the socket descriptor used for comm */
  char *hostname;               /* the hostname of the peer - Tcp sockets */
  int port;                     /* the port number of the peer - Tcp sockets */
  char *path;                   /* the path of the peer - unix sockets */
  int type;                     /* UNIX_SOC or TCP_SOC */
  int port_id;                  /* the port number of the processor - GM port */
  int node_id;                  /* GM node */
  int next_msg;                 /* The type of next MSG to recieve. */
  int args[7];                  /* The header that was received before. */
  _USF_message_header hdr;      /* The header that was received before. */
}
_USF_conn_struct;


#define _USF_RECVPOST 0
#define _USF_SENDPOST 1
#define _USF_FCHUNK   2
#define _USF_MCHUNK   3
#define _USF_LCHUNK   4
#define _USF_ACKCHUNK 5
#define _USF_SHORTMES 6
#define _USF_COMMTERM 7
#define _USF_ACK_GMSEND 18
//Added by Chaney 2002, used for derived datatype
//---
#define _USF_DD_SINGLE_COND      8
#define _USF_DD_SINGLE_END       9
#define _USF_DD_MULTI_COND_COND 10      //chunk of mid element of structure
#define _USF_DD_MULTI_COND_END  11      //chunk of the last element of structure
#define _USF_DD_MULTI_END_COND  12      //the last chunk of mid element of structure
#define _USF_DD_MULTI_END_END   13      //the last chunk of the last element of structure
#define _USF_DD_ACK             14
//---


#define _USF_RECEIVE 0
#define _USF_SEND 1
#define _USF_PROBE 2

typedef struct _USF_rank_type
{
  int rank;
  int comm;
  int nump;
  int group;
  int old_comm;
}
_USF_rank_type;

typedef struct _USF_group_type
{
  int *rank_global;
  int *rank_old;
  int group_no_local;
  int group_no_old;
  int group_size;
}
_USF_group_type;

typedef struct _USF_header
{
  int source;
  int dest;
  int tag;
  int comm;
  int size;
  int type;
  void *in_buf;
  unsigned int send_seq_num;
}
_USF_header;

extern int _USF_group_max;
extern int _USF_group_ref;
extern int _USF_num_of_group;
extern int _USF_comm_max;
extern int _USF_comm_ref;
extern int _USF_num_of_comm;
extern struct _USF_group_type *_USF_group_list;
extern struct _USF_rank_type *_USF_comm_list;


extern pthread_t _USF_comm_deamon;
extern int *_USF_finalize_list;
extern pthread_mutex_t _USF_mut_sysreq; /* sys_request link list */
extern pthread_mutex_t _USF_mut_req;    /* MPI_Request link list */
extern pthread_mutex_t _USF_mut_block;  /* Blocking communication */
extern pthread_cond_t _USF_cond_block;  /* Blocking Communication */
extern int _USF_s_pipe[2];      /* 0 read, 1 write */
// - Bengin - by Qing Huang
extern int _USF_gm_pipe[2];     /* 0 read, 1 write */// by Qing Huang
extern struct gm_port *local_port;
// - End - by Qing Huang

extern MPI_Request_Int *_USF_synchr_list[_USF_MAXPROCESSORS];
extern _USF_sysreq_node *_USF_sysreq_hashtable[_USF_MAXPROCESSORS];
extern _USF_request_node *_USF_request_hd, *_USF_request_tl;
                                              /* Head and tail of the list */
extern _USF_conn_struct *_USF_conn_list;
extern pool_t *_USF_mpi_request_int_pool;
extern pool_t *_USF_request_node_pool;
extern pool_t *_USF_sysreq_node_pool;

// Assume the 0th rank is the one that initiated the run.
// Assume returns 0 in success
int MPI_Init (int *argc, char ***argv);
int MPI_Finalize (void);
int MPI_Recv (void *buf, int count, MPI_Datatype datatype, int source,
              int tag, MPI_Comm comm, MPI_Status * status);
int MPI_Send (void *buf, int count, MPI_Datatype datatype, int dest,
              int tag, MPI_Comm comm);
int MPI_Ssend (void *buf, int count, MPI_Datatype datatype, int dest,
               int tag, MPI_Comm comm);
int MPI_Rsend (void *buf, int count, MPI_Datatype datatype, int dest,
               int tag, MPI_Comm comm);
int MPI_Comm_rank (MPI_Comm comm, int *rank);
int MPI_Comm_size (MPI_Comm comm, int *size);
int MPI_Get_count (MPI_Status * status, MPI_Datatype datatype, int *count);
int MPI_Irecv (void *buf, int count, MPI_Datatype datatype, int source,
               int tag, MPI_Comm comm, MPI_Request * request);
int MPI_Wait (MPI_Request * request, MPI_Status * status);
int MPI_Test (MPI_Request * request, int *flag, MPI_Status * status);
int MPI_Isend (void *buf, int count, MPI_Datatype datatype, int dest,
               int tag, MPI_Comm comm, MPI_Request * request);
int MPI_Irsend (void *buf, int count, MPI_Datatype datatype, int dest,
                int tag, MPI_Comm comm, MPI_Request * request);
int MPI_Issend (void *buf, int count, MPI_Datatype datatype, int dest,
                int tag, MPI_Comm comm, MPI_Request * request);
int MPI_Iprobe (int source, int tag, MPI_Comm comm,
                int *flag, MPI_Status * status);
int MPI_Probe (int source, int tag, MPI_Comm comm, MPI_Status * status);
int MPI_Waitany (int count, MPI_Request * array_of_requests, int *index,
                 MPI_Status * status);
int MPI_Testany (int count, MPI_Request * array_of_requests, int *index,
                 int *flag, MPI_Status * status);
int MPI_Waitall (int count, MPI_Request * array_of_requests,
                 MPI_Status * array_of_statuses);
int MPI_Testall (int count, MPI_Request * array_of_requests, int *flag,
                 MPI_Status * array_of_statuses);
int MPI_Send_init (void *buf, int count, MPI_Datatype datatype, int dest,
                   int tag, MPI_Comm comm, MPI_Request * request);
int MPI_Rsend_init (void *buf, int count, MPI_Datatype datatype, int dest,
                    int tag, MPI_Comm comm, MPI_Request * request);
int MPI_Ssend_init (void *buf, int count, MPI_Datatype datatype, int dest,
                    int tag, MPI_Comm comm, MPI_Request * request);
int MPI_Recv_init (void *buf, int count, MPI_Datatype datatype, int source,
                   int tag, MPI_Comm comm, MPI_Request * request);
int MPI_Start (MPI_Request * request);
int MPI_Startall (int count, MPI_Request * array_of_requests);
int MPI_Sendrecv (void *sendbuf, int sendcount, MPI_Datatype sendtype,
                  int dest, int sendtag, void *recvbuf, int recvcount,
                  MPI_Datatype recvtype, int source, int recvtag,
                  MPI_Comm comm, MPI_Status * status);
int MPI_Sendrecv_replace (void *buf, int count, MPI_Datatype datatype,
                          int dest, int sendtag, int source, int recvtag,
                          MPI_Comm comm, MPI_Status * status);
int MPI_Barrier (MPI_Comm comm);
int MPI_Bcast (void *buffer, int count, MPI_Datatype datatype,
               int root, MPI_Comm comm);
int MPI_Gather (void *sendbuf, int sendcount, MPI_Datatype sendtype,
                void *recvbuf, int recvcount, MPI_Datatype recvtype,
                int root, MPI_Comm comm);
int MPI_Scatter (void *sendbuf, int sendcount, MPI_Datatype sendtype,
                 void *recvbuf, int recvcount, MPI_Datatype recvtype,
                 int root, MPI_Comm comm);
int MPI_Allgather (void *sendbuf, int sendcount, MPI_Datatype sendtype,
                   void *recvbuf, int recvcount, MPI_Datatype recvtype,
                   MPI_Comm comm);
int MPI_Alltoall (void *sendbuf, int sendcount, MPI_Datatype sendtype,
                  void *recvbuf, int recvcount, MPI_Datatype recvtype,
                  MPI_Comm comm);
int MPI_Reduce (void *, void *, int, MPI_Datatype, MPI_Op, int, MPI_Comm);

//--- Add by Chaney 2002
//--- MPI derived data type modual
int MPI_Type_struct (int, int *, MPI_Aint *, MPI_Datatype *, MPI_Datatype *);
int MPI_Address (void *, MPI_Aint *);
int MPI_Pack (void *pack_data, int in_count, MPI_Datatype datatype,
              void *buffer, int buffer_size, int *position, MPI_Comm comm);
int MPI_Unpack (void *buffer, int size, int *position, void *unpack_data,
                int count, MPI_Datatype datatype, MPI_Comm comm);
int MPI_Type_indexed (int count, int block_lengths[],
                      MPI_Aint displacements[], MPI_Datatype old_type,
                      MPI_Datatype * new_mpi_t);
int MPI_Type_contiguous (int count, MPI_Datatype old_type,
                         MPI_Datatype * new_mpi_t);
int MPI_Type_vector (int count, int block_length, int stride,
                     MPI_Datatype element_type, MPI_Datatype * new_mpi_t);
//--- MPI collective
int MPI_Scatterv (void *sendbuf, int *sendcount, int *displs,
                  MPI_Datatype sendtype, void *recvbuf, int recvcount,
                  MPI_Datatype recvtype, int root, MPI_Comm comm);
int MPI_Gatherv (void *sendbuf, int sendcount, MPI_Datatype sendtype,
                 void *recvbuf, int *recvcount, int *displs,
                 MPI_Datatype recvtype, int root, MPI_Comm comm);
int MPI_Op_create (MPI_User_function *, int, MPI_Op *);
int MPI_op_free (MPI_Op * op);
MPI_User_function *Get_User_function (MPI_Op op);
//---
void _USF_handle_recv_post (int dest, int source, int tag, int comm,
                            void *buf, int size, int multiple);
void _USF_handle_send_init (int dest, int source, int tag, int comm,
                            void *buf, int *size);
void _USF_handle_msg (int source);
void _USF_serve_mpi_requests (int rank, int nump);
void *_USF_comm_thread (void *arg);
_USF_sysreq_node *_USF_sysreq_lookup (int tag, int source, int type,
                                      int comm);

#ifdef _USF_MPI_GM_
void _USF_sysreq_insert (int tag, int source, int type, int comm, int size,
                         void *in_buf, _USF_sysreq_node * value);
#endif
#ifdef _USF_MPI_TCP_
void _USF_sysreq_insert (int tag, int source, int type, int comm, int size,
                         _USF_sysreq_node * value);

#endif

_USF_sysreq_node *_USF_sysreq_remove (int tag, int source, int type,
                                      int comm);
void _USF_sysreq_free ();
_USF_request_node *_USF_request_lookup (int tag, int source, int type,
                                        int comm);
_USF_request_node *_USF_request_lookup_m (int tag, int source, int type,
                                          int comm, int multiple);
_USF_request_node *_USF_request_lookup_p (int tag, int type, int comm,
                                          int posted);
void _USF_request_inshd (_USF_request_node * value);
void _USF_request_instl (_USF_request_node * value);
void _USF_request_remove (_USF_request_node * elem);
void _USF_request_free ();

// - Begin - by Qing Huang
int _USF_GM_send_all (int dest, int size, void *buf, char *err_mes);
//int _USF_GM_recv_all (int source, int size, void* buf, char* err_mes);
// - End - by Qing Huang
int _USF_TCP_send_all (int dest, int size, void *buf, char *err_mes);
int _USF_TCP_recv_all (int source, int size, void *buf, char *err_mes);

int _USF_Error (int wh);
int _USF_mpi_sizeof (MPI_Datatype ty);
int _USF_make_conn_list (char *my_name, int nump);

int _USF_num_bits (int n);
int _USF_make_bm (int n);

int _USF_solve_max (void *, void *, int, MPI_Datatype);
int _USF_solve_min (void *, void *, int, MPI_Datatype);
int _USF_solve_sum (void *, void *, int, MPI_Datatype);
int _USF_solve_prod (void *, void *, int, MPI_Datatype);
int _USF_solve_land (void *, void *, int, MPI_Datatype);
int _USF_solve_band (void *, void *, int, MPI_Datatype);
int _USF_solve_lor (void *, void *, int, MPI_Datatype);
int _USF_solve_bor (void *, void *, int, MPI_Datatype);
int _USF_solve_lxor (void *, void *, int, MPI_Datatype);
int _USF_solve_bxor (void *, void *, int, MPI_Datatype);
int _USF_solve_minloc (void *, void *, int, MPI_Datatype);
int _USF_solve_maxloc (void *, void *, int, MPI_Datatype);


// Modified by Qing Huang 
int MPI_Type_size (MPI_Datatype, int *);
double MPI_Wtime ();
int MPI_Reduce_scatter (void *, void *, int *, MPI_Datatype, MPI_Op,
                        MPI_Comm);
int MPI_Allgatherv (void *, int, MPI_Datatype, void *, int *, int *,
                    MPI_Datatype, MPI_Comm);
int MPI_Comm_group (MPI_Comm, MPI_Group *);
int MPI_Abort (MPI_Comm, int);
int MPI_Group_translate_ranks (MPI_Group, int, int *, MPI_Group, int *);
int MPI_Comm_free (MPI_Comm *);
int MPI_Comm_split (MPI_Comm, int, int, MPI_Comm *);
int MPI_Group_incl (MPI_Group old_group,        // in
                    int new_group_size, // in
                    int ranks_in_old_group[],   // in
                    MPI_Group * new_group       // in
  );

int MPI_Comm_create (MPI_Comm old_comm, // in
                     MPI_Group new_group,       // in
                     MPI_Comm * new_comm        // out
  );
// End of Modification QH  

#endif /* _MPI_H */
