#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<unistd.h>
#include"mpi.h"
#include "debug.h"

/* Be careful about inserting to head or tail in request list! */

#ifdef _USF_MPI_GM_             // addded for GM
extern pthread_mutex_t _USF_mut_gm_pending;
extern int request_pending;
extern int thread_recving;

int recv_call = 0;
int isend_call = 0;

unsigned int
address_hash (char *addr, int size)
{
  register unsigned int key;
  //unsigned int key;
  key = (unsigned int) addr;
  return ((key >> 14) * 2654435761 + size) % MAX_MEM_REG_ENTRY;
}


dma_mem_reg_table *mem_reg_table;

int
buf_not_reg (void *p, int size)
{
  int pos;
  dma_mem_reg_table *next = NULL;

  pos = address_hash (p, size);
  next = mem_reg_table + pos;

  if (mem_reg_table[pos].p == NULL) {
    mem_reg_table[pos].p = p;
    mem_reg_table[pos].size = size;
    mem_reg_table[pos].duplicate_link = NULL;
    return 1;
  } else if (mem_reg_table[pos].p == p && mem_reg_table[pos].size == size) {
    return 0;
  } else {
    while (next->duplicate_link != NULL) {
      if (next->duplicate_link->p == p & next->duplicate_link->size == size) {
        return 0;
      } else
        next = next->duplicate_link;
    }

    next->duplicate_link =
      (dma_mem_reg_table *) malloc (sizeof (dma_mem_reg_table));
    if (next->duplicate_link == NULL) {
      printf ("allocate memory error\n");
      fflush (stdout);
      exit (1);
    }
    next = next->duplicate_link;
    next->p = p;
    next->size = size;
    next->duplicate_link = NULL;
    return 1;
  }

  return 1;

}

#endif // end of new additions

inline void
_USF_SR_CONSTRUCTOR (int if_active, int if_blocking, int comm_type,
                     int peer_type, MPI_Request_Int ** req, void *buf,
                     _USF_request_node ** node, int count,
                     MPI_Datatype datatype, int tag, MPI_Comm comm)
{
  int rank;

  MPI_Comm_rank (comm, &rank);
  (*node) = (_USF_request_node *) pool_take (_USF_request_node_pool);
  (*node)->next = NULL;
  (*node)->back = NULL;
  (*req) = (MPI_Request_Int *) pool_take (_USF_mpi_request_int_pool);
  (*req)->active = if_active;
  (*req)->done = 0;
  (*req)->end_of_transmission = 0;
  (*req)->type = comm_type;
  (*req)->blocking = if_blocking;
  (*req)->posted = 0;
  (*req)->buf = buf;
  (*req)->size = count * _USF_mpi_sizeof (datatype);
#ifdef _USF_MPI_GM_
  if (buf != NULL && (*req)->size != 0 && buf_not_reg (buf, (*req)->size)) {

#ifdef DEBUG
    printf ("rank %d in constructor function to register memory\n", rank);
    fflush (stdout);
#endif

    gm_register_memory (local_port, buf, (*req)->size);
  }
#endif
  (*req)->tag = tag;
  (*req)->peer = peer_type;
  (*req)->nchunk = 0;
  (*req)->comm = comm;
  (*req)->node = (*node);
  (*node)->req = (*req);
}

inline void
_USF_SR_MULTIPLE_DEC (MPI_Request_Int * req, int type)
{
#ifdef _USF_MPI_TCP_
  if (type == _USF_SEND) {
    if (req->size > _USF_MAXDATASIZE)
      req->multiple = 1;
    else
      req->multiple = 0;
  } else {
    req->multiple = 0;
  }
#endif
#ifdef _USF_MPI_GM_
  req->multiple = 0;
#endif

}
inline void
_USF_SR_ADD_TOQ (_USF_request_node * node)
{
  pthread_mutex_lock (&_USF_mut_req);
  _USF_request_instl (node);
  pthread_mutex_unlock (&_USF_mut_req);
}
inline void
_USF_SR_NOTIFY (void)
{
  int rank, nump;
#ifndef SINGLE_THREAD
  // multiple threaded mode, try to notify communication thread
#ifdef _USF_MPI_TCP_
  int pipei = 0;
  write (_USF_s_pipe[1], &pipei, sizeof (int));
#else
  pthread_mutex_lock (&_USF_mut_gm_pending);
  request_pending++;
  pthread_mutex_unlock (&_USF_mut_gm_pending);
#endif

#else
  // single threaded mode 
#ifdef _USF_MPI_TCP_
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  MPI_Comm_size (MPI_COMM_WORLD, &nump);
  _USF_serve_mpi_requests (rank, nump);
#endif

#ifdef _USF_MPI_GM_
  request_pending++;
  thread_recving = 1;
  gm_thread ();
#endif

#endif

}

inline void
_USF_SR_BLOCK (_USF_request_node * node)
{

  int nump;                     /* # of procs in COMM_WORLD */
  int rank;                     /* the rank in COMM_WORLD */


  MPI_Comm_size (MPI_COMM_WORLD, &nump);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);



#ifndef SINGLE_THREAD
  pthread_mutex_lock (&_USF_mut_block);
  if (node->req->end_of_transmission == 0)    
    pthread_cond_wait (&_USF_cond_block, &_USF_mut_block);
  pthread_mutex_unlock (&_USF_mut_block);
#else

#ifdef _USF_MPI_TCP_
  int i;                        /* Loop variable */
  fd_set set;                   /* the set that is used for select */
  int max;                      /* the maximum file descriptor */
  int finished = 0;             /* The loop conditional */
  int nu;


  while (!node->req->end_of_transmission) {


    max = _USF_TCP_prepare_set (&set);
    if ((nu = pselect (max + 1, &set, NULL, NULL, NULL, NULL)) == -1)
      Error (ERF_SELECT);
    DBG (("rank %d, woken up from select! sel %d \n", rank, nu));
    for (i = 0; i < nump; i++) {
      if (i == rank)
        continue;
      if (FD_ISSET (_USF_conn_list[i].sd, &set)) {
        DBG (("rank %d handling message %d start\n", rank, i));
        _USF_handle_msg (i);
        DBG (("rank %d handling message %d end\n", rank, i));
      }
    }
  }
#endif

#ifdef _USF_MPI_GM_
  while (!node->req->end_of_transmission) {
    thread_recving = 1;
    gm_thread ();
  }
#endif

#endif

  pthread_mutex_lock (&_USF_mut_req);
  _USF_request_remove (node);
  pthread_mutex_unlock (&_USF_mut_req);

}


int
MPI_Send (void *buf, int count, MPI_Datatype datatype, int dest,
          int tag, MPI_Comm comm)
{
  MPI_Request_Int *req;
  _USF_request_node *node;
  int i;
  _USF_MPI_struct_type *stype;
  int rank;

  if (datatype < MPI_Primitive_Type) {
    if (_USF_check_comm (comm, &dest)) {
      return 1;
    }
    _USF_SR_CONSTRUCTOR (1, 1, _USF_SEND, dest, &req, buf, &node, count,
                         datatype, tag, comm);
    _USF_SR_MULTIPLE_DEC (req, _USF_SEND);
    _USF_SR_ADD_TOQ (node);
    _USF_SR_NOTIFY ();
    _USF_SR_BLOCK (node);
  } else {
    if (_USF_DDType_Get_Type (datatype, &stype) == MPI_SUCCESS)
      MPI_Comm_rank (comm, &rank);
    for (i = 0; i < stype->count; i++) {
      MPI_Send ((void *) (buf + stype->disp[i]), stype->block_lengths[i],
                stype->typelist[i], dest, tag, comm);
    }
  }
  return 0;
}


int
MPI_Rsend (void *buf, int count, MPI_Datatype datatype, int dest,
           int tag, MPI_Comm comm)
{
  MPI_Request_Int *req;
  _USF_request_node *node;
  _USF_MPI_struct_type *stype;
  int rank;
  int i;

  if (datatype < MPI_Primitive_Type) {
    if (_USF_check_comm (comm, &dest)) {
      return 1;
    }
    _USF_SR_CONSTRUCTOR (1, 1, _USF_SEND, dest, &req, buf, &node, count,
                         datatype, tag, comm);
    _USF_SR_MULTIPLE_DEC (req, _USF_SEND);
    _USF_SR_ADD_TOQ (node);
    _USF_SR_NOTIFY ();
    _USF_SR_BLOCK (node);
  } else {
    if (_USF_DDType_Get_Type (datatype, &stype) == MPI_SUCCESS)
      MPI_Comm_rank (comm, &rank);
    for (i = 0; i < stype->count; i++) {
      MPI_Rsend ((void *) (buf + stype->disp[i]), stype->block_lengths[i],
                 stype->typelist[i], dest, tag, comm);
    }
  }
  return 0;
}

int
MPI_Ssend (void *buf, int count, MPI_Datatype datatype, int dest,
           int tag, MPI_Comm comm)
{
  MPI_Request_Int *req;
  _USF_request_node *node;
  _USF_MPI_struct_type *stype;
  int rank;
  int i;

  if (datatype < MPI_Primitive_Type) {
    if (_USF_check_comm (comm, &dest)) {
      return 1;
    }
    _USF_SR_CONSTRUCTOR (1, 1, _USF_SEND, dest, &req, buf, &node, count,
                         datatype, tag, comm);
    _USF_SR_MULTIPLE_DEC (req, _USF_SEND);
    _USF_SR_ADD_TOQ (node);
    _USF_SR_NOTIFY ();
    _USF_SR_BLOCK (node);
  } else {
    if (_USF_DDType_Get_Type (datatype, &stype) == MPI_SUCCESS)
      MPI_Comm_rank (comm, &rank);
    for (i = 0; i < stype->count; i++) {
      MPI_Ssend ((void *) (buf + stype->disp[i]), stype->block_lengths[i],
                 stype->typelist[i], dest, tag, comm);
    }
  }
  return 0;
}

int
MPI_Isend (void *buf, int count, MPI_Datatype datatype, int dest,
           int tag, MPI_Comm comm, MPI_Request * request)
{
  _USF_request_node *node;
  _USF_MPI_struct_type *stype;
  int rank;
  int i;


  if (datatype < MPI_Primitive_Type) {
    if (_USF_check_comm (comm, &dest)) {
      return 1;
    }
    _USF_SR_CONSTRUCTOR (1, 0, _USF_SEND, dest, request, buf, &node, count,
                         datatype, tag, comm);
    _USF_SR_MULTIPLE_DEC ((*request), _USF_SEND);
    _USF_SR_ADD_TOQ (node);
    _USF_SR_NOTIFY ();
  } else {
    if (_USF_DDType_Get_Type (datatype, &stype) == MPI_SUCCESS)
      MPI_Comm_rank (comm, &rank);
    for (i = 0; i < stype->count; i++) {
      MPI_Isend ((void *) (buf + stype->disp[i]), stype->block_lengths[i],
                 stype->typelist[i], dest, tag, comm, request);
    }
  }
  return 0;
}

int
MPI_Irsend (void *buf, int count, MPI_Datatype datatype, int dest,
            int tag, MPI_Comm comm, MPI_Request * request)
{
  _USF_request_node *node;
  _USF_MPI_struct_type *stype;
  int rank;
  int i;

  if (datatype < MPI_Primitive_Type) {
    if (_USF_check_comm (comm, &dest)) {
      return 1;
    }
    _USF_SR_CONSTRUCTOR (1, 0, _USF_SEND, dest, request, buf, &node, count,
                         datatype, tag, comm);
    _USF_SR_MULTIPLE_DEC ((*request), _USF_SEND);
    _USF_SR_ADD_TOQ (node);
    _USF_SR_NOTIFY ();
  } else {
    if (_USF_DDType_Get_Type (datatype, &stype) == MPI_SUCCESS)
      MPI_Comm_rank (comm, &rank);
    for (i = 0; i < stype->count; i++) {
      MPI_Irsend ((void *) (buf + stype->disp[i]), stype->block_lengths[i],
                  stype->typelist[i], dest, tag, comm, request);
    }
  }
  return 0;
}

int
MPI_Issend (void *buf, int count, MPI_Datatype datatype, int dest,
            int tag, MPI_Comm comm, MPI_Request * request)
{
  _USF_request_node *node;
  _USF_MPI_struct_type *stype;
  int rank;
  int i;

  if (datatype < MPI_Primitive_Type) {
    if (_USF_check_comm (comm, &dest)) {
      return 1;
    }
    _USF_SR_CONSTRUCTOR (1, 0, _USF_SEND, dest, request, buf, &node, count,
                         datatype, tag, comm);
    _USF_SR_MULTIPLE_DEC ((*request), _USF_SEND);
    _USF_SR_ADD_TOQ (node);
    _USF_SR_NOTIFY ();
  } else {
    if (_USF_DDType_Get_Type (datatype, &stype) == MPI_SUCCESS)
      MPI_Comm_rank (comm, &rank);
    for (i = 0; i < stype->count; i++) {
      MPI_Issend ((void *) (buf + stype->disp[i]), stype->block_lengths[i],
                  stype->typelist[i], dest, tag, comm, request);
    }
  }
  return 0;
}

int
MPI_Send_init (void *buf, int count, MPI_Datatype datatype, int dest,
               int tag, MPI_Comm comm, MPI_Request * request)
{
  _USF_request_node *node;

  if (_USF_check_comm (comm, &dest)) {
    return 1;
  }
  _USF_SR_CONSTRUCTOR (0, 0, _USF_SEND, dest, request, buf, &node, count,
                       datatype, tag, comm);
  _USF_SR_MULTIPLE_DEC ((*request), _USF_SEND);
  _USF_SR_ADD_TOQ (node);
  return 0;
}

int
MPI_Rsend_init (void *buf, int count, MPI_Datatype datatype, int dest,
                int tag, MPI_Comm comm, MPI_Request * request)
{
  _USF_request_node *node;

  if (_USF_check_comm (comm, &dest)) {
    return 1;
  }
  _USF_SR_CONSTRUCTOR (0, 0, _USF_SEND, dest, request, buf, &node, count,
                       datatype, tag, comm);
  _USF_SR_MULTIPLE_DEC ((*request), _USF_SEND);
  _USF_SR_ADD_TOQ (node);
  return 0;
}

int
MPI_Ssend_init (void *buf, int count, MPI_Datatype datatype, int dest,
                int tag, MPI_Comm comm, MPI_Request * request)
{
  _USF_request_node *node;

  if (_USF_check_comm (comm, &dest)) {
    return 1;
  }
  _USF_SR_CONSTRUCTOR (0, 0, _USF_SEND, dest, request, buf, &node, count,
                       datatype, tag, comm);
  _USF_SR_MULTIPLE_DEC ((*request), _USF_SEND);
  _USF_SR_ADD_TOQ (node);
  return 0;
}

int
MPI_Recv_init (void *buf, int count, MPI_Datatype datatype, int source,
               int tag, MPI_Comm comm, MPI_Request * request)
{
  _USF_request_node *node;

  if (_USF_check_comm (comm, &source)) {
    return 1;
  }
  _USF_SR_CONSTRUCTOR (0, 0, _USF_RECEIVE, source, request, buf, &node,
                       count, datatype, tag, comm);
  _USF_SR_MULTIPLE_DEC ((*request), _USF_RECEIVE);
  _USF_SR_ADD_TOQ (node);
  return 0;
}

int
MPI_Irecv (void *buf, int count, MPI_Datatype datatype, int source,
           int tag, MPI_Comm comm, MPI_Request * request)
{
  _USF_request_node *node;
  int i;
  _USF_MPI_struct_type *stype;

  if (datatype < MPI_Primitive_Type) {
    if (_USF_check_comm (comm, &source)) {
      return 1;
    }
    _USF_SR_CONSTRUCTOR (1, 0, _USF_RECEIVE, source, request, buf, &node,
                         count, datatype, tag, comm);
    _USF_SR_MULTIPLE_DEC ((*request), _USF_RECEIVE);
    _USF_SR_ADD_TOQ (node);
    _USF_SR_NOTIFY ();
  } else {
    _USF_DDType_Get_Type (datatype, &stype);
    for (i = 0; i < stype->count; i++) {
      MPI_Irecv ((void *) buf + stype->disp[i], stype->block_lengths[i],
                 stype->typelist[i], source, tag, comm, request);
    }
  }
  return 0;
}


int
MPI_Recv (void *buf, int count, MPI_Datatype datatype, int source,
          int tag, MPI_Comm comm, MPI_Status * status)
{
  MPI_Request_Int *req;
  _USF_request_node *node;
  int i;

  int size;

  _USF_MPI_struct_type *stype;
  if (datatype < MPI_Primitive_Type) {
    if (_USF_check_comm (comm, &source)) {
      return 1;
    }
    _USF_SR_CONSTRUCTOR (1, 1, _USF_RECEIVE, source, &req, buf, &node,
                         count, datatype, tag, comm);
    _USF_SR_MULTIPLE_DEC (req, _USF_RECEIVE);
    _USF_SR_ADD_TOQ (node);
    _USF_SR_NOTIFY ();
    _USF_SR_BLOCK (node);
  } else {
    _USF_DDType_Get_Type (datatype, &stype);
    for (i = 0; i < stype->count; i++) {
      MPI_Recv ((void *) buf + stype->disp[i], stype->block_lengths[i],
                stype->typelist[i], source, tag, comm, status);
    }
  }
  return 0;
}
