#include<stdio.h>
#include<stdlib.h>
#include<sys/types.h>
#include<unistd.h>
#include"errno.h"
#include"mpi.h"

#define _USF_SOLVER_MAX_EQ(type1)                 \
      for(i=0;i<count;i++) {                      \
              if (p2.type1[i] > p1.type1[i])      \
                      p1.type1[i] = p2.type1[i];  \
      }
#define _USF_SOLVER_MIN_EQ(type1)                 \
      for(i=0;i<count;i++) {                      \
              if (p2.type1[i] < p1.type1[i])      \
                      p1.type1[i] = p2.type1[i];  \
      }
#define _USF_SOLVER_SUM_EQ(type1)                 \
      for(i=0;i<count;i++) {                      \
              p1.type1[i] += p2.type1[i];         \
      }
#define _USF_SOLVER_PROD_EQ(type1)                \
      for(i=0;i<count;i++) {                      \
              p1.type1[i] *= p2.type1[i];         \
      }
#define _USF_SOLVER_LAND_EQ(type1)                          \
      for(i=0;i<count;i++) {                                \
              p1.type1[i] = (p1.type1[i] && p2.type1[i]);   \
      }
#define _USF_SOLVER_BAND_EQ(type1)                         \
      for(i=0;i<count;i++) {                               \
              p1.type1[i] = (p1.type1[i] & p2.type1[i]);   \
      }
#define _USF_SOLVER_LOR_EQ(type1)                           \
      for(i=0;i<count;i++) {                                \
              p1.type1[i] = (p1.type1[i] || p2.type1[i]);   \
      }
#define _USF_SOLVER_BOR_EQ(type1)                          \
      for(i=0;i<count;i++) {                               \
              p1.type1[i] = (p1.type1[i] | p2.type1[i]);   \
      }
#define _USF_SOLVER_LXOR_EQ(type1)                            \
      for(i=0;i<count;i++) {                                  \
              p1.type1[i] = (p1.type1[i] && p2.type1[i]) ||   \
                         ((!p1.type1[i]) && (!p2.type1[i]));  \
       }
#define _USF_SOLVER_BXOR_EQ(type1)                 \
      for(i=0;i<count;i++) {                       \
              p1.type1[i] = (p1.type1[i] ^ p2.type1[i]);   \
      }
#define _USF_SOLVER_MINLOC_EQ(type1)                 \
      for(i=0;i<count;i++) {                      \
              if (p2.type1[i] > p1.type1[i])      \
                      p1.type1[i] = p2.type1[i];  \
                      }
#define _USF_SOLVER_MAXLOC_EQ(type1)                 \
      for(i=0;i<count;i++) {                      \
              if (p2.type1[i] > p1.type1[i])      \
                      p1.type1[i] = p2.type1[i];  \
                      }


#define _USF_SOLVER(type1, type2, equation)   \
      p1.type1 = type2 hostbuf;                   \
      p2.type1 = type2 peerbuf;                   \
      equation(type1);                            \



typedef union _USF_reduce_union
{

  short *shortp;
  int *intp;
  long *longp;
  unsigned short *ushortp;
  unsigned *usignedp;
  unsigned long *ulongp;
  float *floatp;
  double *doublep;
  long double *ldoublep;
  long long *llp;
}
_USF_reduce_union;

#define _USF_SOLVER_GENERAL(funname, solver_eq)                                  \
int funname (void *hostbuf, void *peerbuf, int count,                            \
                    MPI_Datatype datatype)                                       \
{                                                                                \
        int i;                                                                   \
        _USF_reduce_union p1, p2;                                                \
        if(datatype > MPI_LONG_DOUBLE_INT) {                                     \
                printf("Unexpected Datatype!\n");                                \
                exit(1);                                                         \
        }                                                                        \
        switch(datatype) {                                                       \
        case MPI_SHORT:                                                          \
                _USF_SOLVER(shortp, (short *), solver_eq);                       \
                break;                                                           \
        case MPI_INT:                                                            \
                _USF_SOLVER(intp, (int *), solver_eq);                           \
                break;                                                           \
        case MPI_LONG:                                                           \
                _USF_SOLVER(longp, (long *), solver_eq);                         \
                break;                                                           \
        case MPI_UNSIGNED_SHORT:                                                 \
                _USF_SOLVER(ushortp, (unsigned short*), solver_eq);              \
                break;                                                           \
        case MPI_UNSIGNED:                                                       \
                _USF_SOLVER(usignedp, (unsigned *), solver_eq);                  \
                break;                                                           \
        case MPI_UNSIGNED_LONG:                                                  \
                _USF_SOLVER(ulongp, (unsigned long*), solver_eq);                \
                break;                                                           \
        case MPI_FLOAT:                                                          \
                _USF_BINARY_GEN(_USF_SOLVER(floatp, (float *), solver_eq));      \
                break;                                                           \
        case MPI_DOUBLE:                                                         \
                _USF_BINARY_GEN(_USF_SOLVER(doublep, (double *), solver_eq));    \
                break;                                                           \
        case MPI_LONG_DOUBLE:                                                    \
                _USF_BINARY_GEN(_USF_SOLVER(ldoublep, (long double*), solver_eq));\
                break;                                                           \
        case MPI_LONG_LONG:                                                      \
                _USF_SOLVER(llp, (long long*), solver_eq);                       \
                break;                                                           \
        default:                                                                 \
                printf("This operation isn't defined for this datatype!\n");     \
                exit(1);                                                         \
        }                                                                        \
        return 0;                                                                \
}

#define _USF_BINARY_GEN(x) x
_USF_SOLVER_GENERAL (_USF_solve_max, _USF_SOLVER_MAX_EQ);
_USF_SOLVER_GENERAL (_USF_solve_min, _USF_SOLVER_MIN_EQ);
_USF_SOLVER_GENERAL (_USF_solve_sum, _USF_SOLVER_SUM_EQ);
_USF_SOLVER_GENERAL (_USF_solve_prod, _USF_SOLVER_PROD_EQ);
_USF_SOLVER_GENERAL (_USF_solve_land, _USF_SOLVER_LAND_EQ);
_USF_SOLVER_GENERAL (_USF_solve_lor, _USF_SOLVER_LOR_EQ);
_USF_SOLVER_GENERAL (_USF_solve_lxor, _USF_SOLVER_LXOR_EQ);

#undef _USF_BINARY_GEN(x)
#define _USF_BINARY_GEN(x)
_USF_SOLVER_GENERAL (_USF_solve_bxor, _USF_SOLVER_BXOR_EQ);
_USF_SOLVER_GENERAL (_USF_solve_band, _USF_SOLVER_BAND_EQ);
_USF_SOLVER_GENERAL (_USF_solve_bor, _USF_SOLVER_BOR_EQ);

typedef union _USF_reduce_double_union
{

  _USF_FLOAT_INT *ufip;
  _USF_DOUBLE_INT *udip;
  _USF_LONG_INT *ulip;
  _USF_SHORT_INT *usip;
  _USF_2INT *uiip;
  _USF_LONG_DOUBLE_INT *uldip;
}
_USF_reduce_double_union;


int
_USF_solve_2loc (void *hostbuf, void *peerbuf, int count,
                 MPI_Datatype datatype, int min)
{
  int i;

  _USF_reduce_double_union p1, p2;

  if (datatype > MPI_LONG_DOUBLE_INT) {
    printf ("Unexpected Datatype!\n");
    exit (1);
  }
  switch (datatype) {
  case MPI_FLOAT_INT:
    p1.ufip = (_USF_FLOAT_INT *) hostbuf;
    p2.ufip = (_USF_FLOAT_INT *) peerbuf;
    for (i = 0; i < count; i++) {
      if (min) {
        if (p2.ufip[i].fst < p1.ufip[i].fst) {
          p1.ufip[i].fst = p2.ufip[i].fst;
          p1.ufip[i].sec = p2.ufip[i].sec;
        }
      } else {
        if (p2.ufip[i].fst > p1.ufip[i].fst) {
          p1.ufip[i].fst = p2.ufip[i].fst;
          p1.ufip[i].sec = p2.ufip[i].sec;
        }
      }
    }
    break;
  case MPI_DOUBLE_INT:
    p1.udip = (_USF_DOUBLE_INT *) hostbuf;
    p2.udip = (_USF_DOUBLE_INT *) peerbuf;
    for (i = 0; i < count; i++) {
      if (min) {
        if (p2.udip[i].fst < p1.udip[i].fst) {
          p1.udip[i].fst = p2.udip[i].fst;
          p1.udip[i].sec = p2.udip[i].sec;
        }
      } else {
        if (p2.udip[i].fst > p1.udip[i].fst) {
          p1.udip[i].fst = p2.udip[i].fst;
          p1.udip[i].sec = p2.udip[i].sec;
        }
      }
    }
    break;
  case MPI_LONG_INT:
    p1.ulip = (_USF_LONG_INT *) hostbuf;
    p2.ulip = (_USF_LONG_INT *) peerbuf;
    for (i = 0; i < count; i++) {
      if (min) {
        if (p2.ulip[i].fst < p1.ulip[i].fst) {
          p1.ulip[i].fst = p2.ulip[i].fst;
          p1.ulip[i].sec = p2.ulip[i].sec;
        }
      } else {
        if (p2.ulip[i].fst > p1.ulip[i].fst) {
          p1.ulip[i].fst = p2.ulip[i].fst;
          p1.ulip[i].sec = p2.ulip[i].sec;
        }
      }
    }
    break;
  case MPI_SHORT_INT:
    p1.usip = (_USF_SHORT_INT *) hostbuf;
    p2.usip = (_USF_SHORT_INT *) peerbuf;
    for (i = 0; i < count; i++) {
      if (min) {
        if (p2.usip[i].fst < p1.usip[i].fst) {
          p1.usip[i].fst = p2.usip[i].fst;
          p1.usip[i].sec = p2.usip[i].sec;
        }
      } else {
        if (p2.usip[i].fst > p1.usip[i].fst) {
          p1.usip[i].fst = p2.usip[i].fst;
          p1.usip[i].sec = p2.usip[i].sec;
        }
      }
    }
    break;
  case MPI_2INT:
    p1.uiip = (_USF_2INT *) hostbuf;
    p2.uiip = (_USF_2INT *) peerbuf;
    for (i = 0; i < count; i++) {
      if (min) {
        if (p2.uiip[i].fst < p1.uiip[i].fst) {
          p1.uiip[i].fst = p2.uiip[i].fst;
          p1.uiip[i].sec = p2.uiip[i].sec;
        }
      } else {
        if (p2.uiip[i].fst > p1.uiip[i].fst) {
          p1.uiip[i].fst = p2.uiip[i].fst;
          p1.uiip[i].sec = p2.uiip[i].sec;
        }
      }
    }
    break;
  case MPI_LONG_DOUBLE_INT:
    p1.uldip = (_USF_LONG_DOUBLE_INT *) hostbuf;
    p2.uldip = (_USF_LONG_DOUBLE_INT *) peerbuf;
    for (i = 0; i < count; i++) {
      if (min) {
        if (p2.uldip[i].fst < p1.uldip[i].fst) {
          p1.uldip[i].fst = p2.uldip[i].fst;
          p1.uldip[i].sec = p2.uldip[i].sec;
        }
      } else {
        if (p2.uldip[i].fst > p1.uldip[i].fst) {
          p1.uldip[i].fst = p2.uldip[i].fst;
          p1.uldip[i].sec = p2.uldip[i].sec;
        }
      }
    }
    break;
  default:
    printf ("This operation isn't defined for this datatype!\n");
    exit (1);


  }

  return 0;
}

// tag -15 is used for MPI_REDUCE
int
MPI_Reduce (void *sendbuf, void *recvbuf, int count,
            MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm)
{
  int nump, rank;
  int bm, nb, cb;
  int partner;
  MPI_Status status;
  void *buf1, *buf2;
  MPI_User_function *userop;


  MPI_Comm_size (comm, &nump);
  MPI_Comm_rank (comm, &rank);

  if (rank == -1) {
    printf ("Wrong communicator in calling MPI_Reduce\n");
    fflush (stdout);
    return 1;
  }

  nb = _USF_num_bits (nump - 1);
  bm = _USF_make_bm (nb);

  if (rank == 0)
    rank = root;
  else if (rank == root)
    rank = 0;
  for (cb = 1; cb <= bm; cb <<= 1) {
    partner = rank ^ cb;

    if (rank < partner) {
      if (partner == 0)
        partner = root;
      else if (partner == root)
        partner = 0;
      if (partner >= nump)
        continue;
      MPI_Recv (recvbuf, count, datatype, partner, -15, comm, &status);
      if ((rank == 0) && (cb == bm)) {
        buf1 = recvbuf;
        buf2 = sendbuf;
      } else {
        buf1 = sendbuf;
        buf2 = recvbuf;
      }
      switch (op) {
      case MPI_MAX:
        _USF_solve_max (buf1, buf2, count, datatype);
        break;
      case MPI_MIN:
        _USF_solve_min (buf1, buf2, count, datatype);
        break;
      case MPI_SUM:
        _USF_solve_sum (buf1, buf2, count, datatype);
        break;
      case MPI_PROD:
        _USF_solve_prod (buf1, buf2, count, datatype);
        break;
      case MPI_LAND:
        _USF_solve_land (buf1, buf2, count, datatype);
        break;
      case MPI_BAND:
        _USF_solve_band (buf1, buf2, count, datatype);
        break;
      case MPI_LOR:
        _USF_solve_lor (buf1, buf2, count, datatype);
        break;
      case MPI_BOR:
        _USF_solve_bor (buf1, buf2, count, datatype);
        break;
      case MPI_LXOR:
        _USF_solve_lxor (buf1, buf2, count, datatype);
        break;
      case MPI_BXOR:
        _USF_solve_bxor (buf1, buf2, count, datatype);
        break;
      case MPI_MINLOC:
        _USF_solve_2loc (buf1, buf2, count, datatype, 1);
        break;
      case MPI_MAXLOC:
        _USF_solve_2loc (buf1, buf2, count, datatype, 0);
        break;
      default:
        if (op >= 120) {
          //This is a user defined opreation
          userop = Get_User_function (op);
          if (userop == NULL) {
            printf ("This operation isn't defined!\n");
            exit (1);
          } else {
            userop (buf2, buf1, &count, &datatype);
          }
        } else {
          printf ("This operation isn't defined!\n");
          exit (1);
        }
      }
    } else {
      if (partner == 0)
        partner = root;
      else if (partner == root)
        partner = 0;
      MPI_Send (sendbuf, count, datatype, partner, -15, comm);
      break;
    }
  }

  return 0;
}
