#include<stdio.h>
#include "mpi.h"

_USF_request_node *_USF_request_hd, *_USF_request_tl;


//Looking for local send _USF_SEND request
_USF_request_node *
_USF_request_lookup (int tag, int peer, int type, int comm)
{
  _USF_request_node *node = _USF_request_hd;

  while (node != NULL) {
#ifdef DEBUG
    printf (" ---> t %d pe %d ty %d c %d\n", tag, peer, type, comm);
    printf (" ---> t %d pe %d ty %d c %d\n", node->req->tag, node->req->peer,
            node->req->type, node->req->comm);
    fflush (stdout);
#endif
    if (((tag == node->req->tag) ||
         ((tag == MPI_ANY_TAG) && (tag >= -1))) &&
        (peer == node->req->peer) && (type == node->req->type)
        && (comm == node->req->comm) && (node->req->done == 0)
        && (node->req->active))
      return node;
    node = node->next;
  }
  return node;
}

//Looking for multiple local request
_USF_request_node *
_USF_request_lookup_m (int tag, int peer, int type, int comm, int multiple)
{
  _USF_request_node *node = _USF_request_hd;

  while (node != NULL) {

#ifdef DEBUG
    printf ("m---> t %d pe %d ty %d c %d m %d\n", tag, peer, type, comm,
            multiple);
    printf ("m---> t %d pe %d ty %d c %d m %d a %d d %d\n", node->req->tag,
            node->req->peer, node->req->type, node->req->comm,
            node->req->multiple, (node->req->active), node->req->done);
    fflush (stdout);
#endif

    if (type == _USF_SEND) {    // Modified by Qing Huang

      if (((tag == node->req->tag) || (tag == MPI_ANY_TAG)) &&
          (peer == node->req->peer) && (type == node->req->type)
          && (comm == node->req->comm) && (node->req->done == 0)
          && (multiple == node->req->multiple) && (node->req->active))
        return node;

    } else if (type == _USF_RECEIVE) {
      if (((tag == node->req->tag) || (node->req->tag == MPI_ANY_TAG))
          && ((peer == node->req->peer)
              || (node->req->peer == MPI_ANY_SOURCE))
          // Modified by Qing Hunag
          && (type == node->req->type)
          && (comm == node->req->comm) && (node->req->done == 0)
          && (multiple == node->req->multiple) && (node->req->active))
        return node;
    }
    node = node->next;
  }
  return node;
}


//Looking for local send _USF_RECEIVE or _USF_PROBE request
_USF_request_node *
_USF_request_lookup_p (int tag, int type, int comm, int posted)
{
  _USF_request_node *node = _USF_request_hd;

  while (node != NULL) {
#ifdef DEBUG
    printf ("p---> t %d ty %d c %d p %d \n", tag, type, comm, posted);
    printf ("p---> t %d ty %d c %d p %d\n", node->req->tag, node->req->type,
            node->req->comm, node->req->posted);
    fflush (stdout);
#endif
    if (((tag == node->req->tag) ||
         ((node->req->tag == MPI_ANY_TAG) && (node->req->tag >= -1))
         || ((tag == MPI_ANY_TAG) && (tag >= -1))) &&
        (type == node->req->type) && (comm == node->req->comm) &&
        (posted == node->req->posted) && (node->req->done == 0)
        && (node->req->multiple == 0) && (node->req->active))
      //may not need multiple here
      return node;
    node = node->next;
  }
  return node;
}

void
_USF_request_inshd (_USF_request_node * value)
{
  value->next = NULL;
  value->back = NULL;
  if (_USF_request_hd == NULL) {
    _USF_request_hd = value;
    _USF_request_tl = value;
  } else {
    _USF_request_hd->back = value;
    value->next = _USF_request_hd;
    _USF_request_hd = value;
  }
  return;
}

void
_USF_request_instl (_USF_request_node * value)
{
  value->next = NULL;
  value->back = NULL;
  if (_USF_request_hd == NULL) {
    _USF_request_hd = value;
    _USF_request_tl = value;
  } else {
    _USF_request_tl->next = value;
    value->back = _USF_request_tl;
    _USF_request_tl = value;
  }
  return;
}

void
_USF_request_remove (_USF_request_node * elem)
{
  if (_USF_request_hd == NULL || elem == NULL) {        
    // Shouldn't happen in a correct program. 
    printf ("Cannot be removed, request queue is empty.\n");
    fflush (stdout);
    return;
  }
  if (_USF_request_hd->next == NULL) {  /* One element */
    if (_USF_request_hd != elem) {
      printf
        ("Request is wrong, the node doesn't belong to the reques queue.\n");
      fflush (stdout);
      return;                   /* Shouldn't happen in a correct program. */
    }
    _USF_request_hd = NULL;
    _USF_request_tl = NULL;
  } else if (_USF_request_hd == elem) {
    elem->next->back = NULL;
    _USF_request_hd = elem->next;
  } else if (_USF_request_tl == elem) {
    elem->back->next = NULL;
    _USF_request_tl = elem->back;
  } else {
    elem->next->back = elem->back;
    elem->back->next = elem->next;
  }
  if (elem != NULL) {           /* Free the element */
    if (elem->req != NULL)
      pool_return (_USF_mpi_request_int_pool, (pool_item *) elem->req);
    pool_return (_USF_request_node_pool, (pool_item *) elem);
  }
  return;
}

void
_USF_request_free ()
{
  int i;
  _USF_request_node *curr = _USF_request_hd;
  _USF_request_node *elem;


  _USF_request_hd = NULL;
  _USF_request_tl = NULL;
  while (curr != NULL) {
    elem = curr;
    curr = curr->next;
    if (elem != NULL) {         /* Free the element */
      if (elem->req != NULL)
        pool_return (_USF_mpi_request_int_pool, (pool_item *) elem->req);
      pool_return (_USF_request_node_pool, (pool_item *) elem);
    }
  }
  return;
}
