// MPI user defined reduce function implementation
// Created by Chaney 2002
// Uniersity of San Francisco

#include <string.h>
#include <stdlib.h>
#include "mpi.h"


typedef struct StructUserFunc
{
  int idx;
  int commute;
  MPI_User_function *ptr2userfunc;
  struct StructUserFunc *next;
}
USF_User_Func_Node;

USF_User_Func_Node *head_userfunc = NULL;
//Set start number of user define op
int User_Op_Index = 120;

//Create user define operation
int
MPI_Op_create (MPI_User_function * userfunc, int commute, MPI_Op * op)
{
  USF_User_Func_Node *funcnode;

  funcnode = (USF_User_Func_Node *) malloc (sizeof (USF_User_Func_Node));
  funcnode->idx = User_Op_Index;
  *op = User_Op_Index;
  User_Op_Index++;
  funcnode->commute = commute;
  funcnode->ptr2userfunc = userfunc;
  funcnode->next = head_userfunc;
  head_userfunc = funcnode;
}

//Create user define operation
int
MPI_op_free (MPI_Op * op)
{
  USF_User_Func_Node *ptr, *pre;

  ptr = head_userfunc;
  pre = ptr;
  while (ptr != NULL) {
    if (ptr->idx == *op) {
      *op = MPI_OP_NULL;
      if (pre != ptr) {
        pre->next = ptr->next;
      } else {
        head_userfunc = ptr->next;
      }
      free (ptr);
      return MPI_SUCCESS;
    }
    pre = ptr;
    ptr = ptr->next;
  }
  return -1;
}

//Get user define operation (function pointer)
MPI_User_function *
Get_User_function (MPI_Op op)
{
  USF_User_Func_Node *ptr;

  ptr = head_userfunc;
  while (ptr != NULL) {
    if (ptr->idx == op) {
      return ptr->ptr2userfunc;
    }
    ptr = ptr->next;
  }
  return NULL;
}
